Package: adbcdrivermanager
Title: 'Arrow' Database Connectivity ('ADBC') Driver Manager
Version: 0.24.0-2
Authors@R: c(
      person("Dewey", "Dunnington", , "dewey@dunnington.ca", role = c("aut", "cre"),
             comment = c(ORCID = "0000-0002-9415-4582")),
      person("Apache Arrow", email = "dev@arrow.apache.org", role = c("aut", "cph")),
      person("Apache Software Foundation", email = "dev@arrow.apache.org", role = c("cph"))
    )
Description: Provides a developer-facing interface to 'Arrow' Database
  Connectivity ('ADBC') for the purposes of driver development, driver
  testing, and building high-level database interfaces for users. 'ADBC'
  <https://arrow.apache.org/adbc/> is an API standard for database access
  libraries that uses 'Arrow' for result sets and query parameters.
License: Apache License (>= 2)
Encoding: UTF-8
Suggests: testthat (>= 3.0.0), withr
Config/testthat/edition: 3
Config/build/bootstrap: TRUE
URL: https://arrow.apache.org/adbc/current/r/adbcdrivermanager/,
        https://github.com/apache/arrow-adbc
BugReports: https://github.com/apache/arrow-adbc/issues
Imports: nanoarrow (>= 0.3.0)
Config/roxygen2/version: 8.0.0
NeedsCompilation: yes
Packaged: 2026-08-23 02:42:08 UTC; dewey
Author: Dewey Dunnington [aut, cre] (ORCID:
    <https://orcid.org/0000-0002-9415-4582>),
  Apache Arrow [aut, cph],
  Apache Software Foundation [cph]
Maintainer: Dewey Dunnington <dewey@dunnington.ca>
Repository: CRAN
Date/Publication: 2026-08-23 07:40:02 UTC
Built: R 4.6.1; aarch64-apple-darwin23; 2026-08-23 08:59:52 UTC; unix
Archs: adbcdrivermanager.so.dSYM

# Generated by roxygen2: do not edit by hand

S3method("$",adbc_error)
S3method("$",adbc_xptr)
S3method("$<-",adbc_xptr)
S3method("[[",adbc_error)
S3method("[[",adbc_xptr)
S3method("[[<-",adbc_xptr)
S3method(adbc_connection_init,adbc_database_log)
S3method(adbc_connection_init,adbc_database_monkey)
S3method(adbc_connection_init,default)
S3method(adbc_connection_quote_identifier,default)
S3method(adbc_connection_quote_string,default)
S3method(adbc_database_init,"NULL")
S3method(adbc_database_init,adbc_driver)
S3method(adbc_database_init,adbc_driver_log)
S3method(adbc_database_init,adbc_driver_monkey)
S3method(adbc_database_init,character)
S3method(adbc_statement_init,adbc_connection_log)
S3method(adbc_statement_init,adbc_connection_monkey)
S3method(adbc_statement_init,default)
S3method(execute_adbc,default)
S3method(format,adbc_xptr)
S3method(length,adbc_error)
S3method(length,adbc_xptr)
S3method(names,adbc_error)
S3method(names,adbc_xptr)
S3method(print,adbc_driver)
S3method(print,adbc_error)
S3method(print,adbc_xptr)
S3method(read_adbc,default)
S3method(str,adbc_driver)
S3method(str,adbc_error)
S3method(str,adbc_xptr)
S3method(write_adbc,default)
export(adbc_connection_cancel)
export(adbc_connection_commit)
export(adbc_connection_get_info)
export(adbc_connection_get_objects)
export(adbc_connection_get_option)
export(adbc_connection_get_option_bytes)
export(adbc_connection_get_option_double)
export(adbc_connection_get_option_int)
export(adbc_connection_get_statistic_names)
export(adbc_connection_get_statistics)
export(adbc_connection_get_table_schema)
export(adbc_connection_get_table_types)
export(adbc_connection_init)
export(adbc_connection_init_default)
export(adbc_connection_join)
export(adbc_connection_quote_identifier)
export(adbc_connection_quote_string)
export(adbc_connection_read_partition)
export(adbc_connection_release)
export(adbc_connection_rollback)
export(adbc_connection_set_options)
export(adbc_database_get_option)
export(adbc_database_get_option_bytes)
export(adbc_database_get_option_double)
export(adbc_database_get_option_int)
export(adbc_database_init)
export(adbc_database_init_default)
export(adbc_database_release)
export(adbc_database_set_options)
export(adbc_driver)
export(adbc_driver_load)
export(adbc_driver_log)
export(adbc_driver_monkey)
export(adbc_driver_void)
export(adbc_error_from_array_stream)
export(adbc_statement_bind)
export(adbc_statement_bind_stream)
export(adbc_statement_cancel)
export(adbc_statement_execute_query)
export(adbc_statement_execute_schema)
export(adbc_statement_get_option)
export(adbc_statement_get_option_bytes)
export(adbc_statement_get_option_double)
export(adbc_statement_get_option_int)
export(adbc_statement_get_parameter_schema)
export(adbc_statement_init)
export(adbc_statement_init_default)
export(adbc_statement_join)
export(adbc_statement_prepare)
export(adbc_statement_release)
export(adbc_statement_set_options)
export(adbc_statement_set_sql_query)
export(adbc_statement_set_substrait_plan)
export(adbc_xptr_is_valid)
export(adbc_xptr_move)
export(execute_adbc)
export(local_adbc)
export(read_adbc)
export(with_adbc)
export(write_adbc)
importFrom(utils,str)
useDynLib(adbcdrivermanager, .registration = TRUE)

<!---
  Licensed to the Apache Software Foundation (ASF) under one
  or more contributor license agreements.  See the NOTICE file
  distributed with this work for additional information
  regarding copyright ownership.  The ASF licenses this file
  to you under the Apache License, Version 2.0 (the
  "License"); you may not use this file except in compliance
  with the License.  You may obtain a copy of the License at
    http://www.apache.org/licenses/LICENSE-2.0
  Unless required by applicable law or agreed to in writing,
  software distributed under the License is distributed on an
  "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
  KIND, either express or implied.  See the License for the
  specific language governing permissions and limitations
  under the License.
-->
<!-- README.md is generated from README.Rmd. Please edit that file -->

# adbcdrivermanager

<!-- badges: start -->
<!-- badges: end -->

The goal of adbcdrivermanager is to provide a low-level developer-facing
interface to Arrow Database Connectivity (ADBC) for the purposes of
driver development, testing, and support for user-facing packages that
rely on ADBC drivers.

## Installation

You can install the released version of adbcdrivermanager from
[CRAN](https://cran.r-project.org/) with:

``` r
install.packages("adbcdrivermanager")
```

You can install the development version of adbcdrivermanager from
[GitHub](https://github.com/) with:

``` r
# install.packages("pak")
pak::pak("apache/arrow-adbc/r/adbcdrivermanager")
```

ADBC drivers for R use a relatively new feature of pkgbuild to enable
installation from GitHub via pak. Depending on when you installed pak,
you may need to update its internal version of pkgbuild.

``` r
install.packages("pkgbuild", pak:::private_lib_dir())
pak::cache_clean()
```

## Example

This is a basic example which shows you how to solve a common problem:

``` r
library(adbcdrivermanager)

# Get a reference to a database using a driver. The adbcdrivermanager package
# contains a few drivers useful for illustration and testing.
db <- adbc_database_init(adbcsqlite::adbcsqlite())

# Open a new connection to a database
con <- adbc_connection_init(db)

# Write a table
nycflights13::flights |> write_adbc(con, "flights")

# Issue a query
con |>
  read_adbc("SELECT * from flights") |>
  tibble::as_tibble()
#> # A tibble: 336,776 × 19
#>     year month   day dep_time sched_dep_time dep_delay arr_time sched_arr_time
#>    <dbl> <dbl> <dbl>    <dbl>          <dbl>     <dbl>    <dbl>          <dbl>
#>  1  2013     1     1      517            515         2      830            819
#>  2  2013     1     1      533            529         4      850            830
#>  3  2013     1     1      542            540         2      923            850
#>  4  2013     1     1      544            545        -1     1004           1022
#>  5  2013     1     1      554            600        -6      812            837
#>  6  2013     1     1      554            558        -4      740            728
#>  7  2013     1     1      555            600        -5      913            854
#>  8  2013     1     1      557            600        -3      709            723
#>  9  2013     1     1      557            600        -3      838            846
#> 10  2013     1     1      558            600        -2      753            745
#> # ℹ 336,766 more rows
#> # ℹ 11 more variables: arr_delay <dbl>, carrier <chr>, flight <dbl>,
#> #   tailnum <chr>, origin <chr>, dest <chr>, air_time <dbl>, distance <dbl>,
#> #   hour <dbl>, minute <dbl>, time_hour <chr>

# Clean up!
adbc_connection_release(con)
adbc_database_release(db)
```

One can also interact with the driver manager at a lower level using
`adbc_connection_*()` and `adbc_statement_*()` functions (see reference
documentation for details).

# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

#' @rdname adbc_database_init
#' @export
adbc_database_set_options <- function(database, options) {
  options <- key_value_options(options)
  error <- adbc_allocate_error()
  for (i in seq_along(options)) {
    key <- names(options)[i]
    value <- options[[i]]
    status <- .Call(
      RAdbcDatabaseSetOption,
      database,
      key,
      value,
      error
    )
    stop_for_error(status, error)
  }
  invisible(database)
}

#' @rdname adbc_connection_init
#' @export
adbc_connection_set_options <- function(connection, options) {
  options <- key_value_options(options)
  error <- adbc_allocate_error()
  for (i in seq_along(options)) {
    key <- names(options)[i]
    value <- options[[i]]
    status <- .Call(
      RAdbcConnectionSetOption,
      connection,
      key,
      value,
      error
    )
    stop_for_error(status, error)
  }
  invisible(connection)
}

#' @rdname adbc_statement_init
#' @export
adbc_statement_set_options <- function(statement, options) {
  options <- key_value_options(options)
  error <- adbc_allocate_error()
  for (i in seq_along(options)) {
    key <- names(options)[i]
    value <- options[[i]]
    status <- .Call(
      RAdbcStatementSetOption,
      statement,
      key,
      value,
      error
    )
    stop_for_error(status, error)
  }
  invisible(statement)
}

#' @rdname adbc_database_init
#' @export
adbc_database_get_option <- function(database, option) {
  error <- adbc_allocate_error()
  .Call(RAdbcDatabaseGetOption, database, option, error)
}

#' @rdname adbc_database_init
#' @export
adbc_database_get_option_bytes <- function(database, option) {
  error <- adbc_allocate_error()
  .Call(RAdbcDatabaseGetOptionBytes, database, option, error)
}

#' @rdname adbc_database_init
#' @export
adbc_database_get_option_int <- function(database, option) {
  error <- adbc_allocate_error()
  .Call(RAdbcDatabaseGetOptionInt, database, option, error)
}

#' @rdname adbc_database_init
#' @export
adbc_database_get_option_double <- function(database, option) {
  error <- adbc_allocate_error()
  .Call(RAdbcDatabaseGetOptionDouble, database, option, error)
}


#' @rdname adbc_connection_init
#' @export
adbc_connection_get_option <- function(connection, option) {
  error <- adbc_allocate_error()
  .Call(RAdbcConnectionGetOption, connection, option, error)
}

#' @rdname adbc_connection_init
#' @export
adbc_connection_get_option_bytes <- function(connection, option) {
  error <- adbc_allocate_error()
  .Call(RAdbcConnectionGetOptionBytes, connection, option, error)
}

#' @rdname adbc_connection_init
#' @export
adbc_connection_get_option_int <- function(connection, option) {
  error <- adbc_allocate_error()
  .Call(RAdbcConnectionGetOptionInt, connection, option, error)
}

#' @rdname adbc_connection_init
#' @export
adbc_connection_get_option_double <- function(connection, option) {
  error <- adbc_allocate_error()
  .Call(RAdbcConnectionGetOptionDouble, connection, option, error)
}


#' @rdname adbc_statement_init
#' @export
adbc_statement_get_option <- function(statement, option) {
  error <- adbc_allocate_error()
  .Call(RAdbcStatementGetOption, statement, option, error)
}

#' @rdname adbc_statement_init
#' @export
adbc_statement_get_option_bytes <- function(statement, option) {
  error <- adbc_allocate_error()
  .Call(RAdbcStatementGetOptionBytes, statement, option, error)
}

#' @rdname adbc_statement_init
#' @export
adbc_statement_get_option_int <- function(statement, option) {
  error <- adbc_allocate_error()
  .Call(RAdbcStatementGetOptionInt, statement, option, error)
}

#' @rdname adbc_statement_init
#' @export
adbc_statement_get_option_double <- function(statement, option) {
  error <- adbc_allocate_error()
  .Call(RAdbcStatementGetOptionDouble, statement, option, error)
}

# Ensures that options are a list of bare character, raw, integer, or double
key_value_options <- function(options) {
  options <- as.list(options)

  if (length(options) == 0) {
    names(options) <- character()
  } else if (is.null(names(options))) {
    # OK to have no names, because options could contain a series of
    # adbc_options() objects that will be concatenated
    names(options) <- rep("", length(options))
  }

  out <- vector("list", 10L)
  out_names <- character(10L)
  n_out <- 0L

  for (i in seq_along(options)) {
    key <- names(options)[[i]]
    item <- options[[i]]

    # Skip NULL item
    if (is.null(item)) {
      next
    }

    # Append all items of an existing adbc_options
    if (inherits(item, "adbc_options")) {
      out <- c(out[seq_len(n_out)], item)
      out_names <- c(out_names[seq_len(n_out)], names(item))
      n_out <- n_out + length(item)
      next
    }

    # Otherwise, append a single value (coercing to character if item
    # is an S3 object)
    if (is.object(item)) {
      item <- as.character(item)
    }

    if (identical(key, "") || identical(key, NA_character_)) {
      stop("key/value options must be named")
    }

    n_out <- n_out + 1L
    out_names[n_out] <- key
    if (is.character(item)) {
      out[[n_out]] <- item
    } else if (is.integer(item)) {
      out[[n_out]] <- as.integer(item)
    } else if (is.double(item)) {
      out[[n_out]] <- as.double(item)
    } else if (is.raw(item)) {
      out[[n_out]] <- item
    } else if (is.logical(item)) {
      out[[n_out]] <- tolower(as.character(item))
    } else {
      stop(
        sprintf(
          "Option of type '%s' (key: '%s') not supported",
           typeof(item),
          key
        )
      )
    }
  }

  names(out) <- out_names
  structure(out[seq_len(n_out)], class = "adbc_options")
}

# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

test_that("get/set option can roundtrip string options for database", {
  db <- adbc_database_init(adbc_driver_void())
  expect_error(
    adbc_database_get_option(db, "some_key"),
    class = "adbc_status_not_found"
  )

  adbc_database_set_options(db, list("some_key" = "some value"))
  expect_identical(
    adbc_database_get_option(db, "some_key"),
    "some value"
  )
})

test_that("get/set option can roundtrip bytes options for database", {
  db <- adbc_database_init(adbc_driver_void())
  expect_error(
    adbc_database_get_option_bytes(db, "some_key"),
    class = "adbc_status_not_found"
  )

  adbc_database_set_options(db, list("some_key" = charToRaw("some value")))
  expect_identical(
    adbc_database_get_option_bytes(db, "some_key"),
    charToRaw("some value")
  )
})

test_that("get/set option can roundtrip integer options for database", {
  db <- adbc_database_init(adbc_driver_void())

  expect_error(
    adbc_database_get_option_int(db, "some_key"),
    class = "adbc_status_not_found"
  )

  adbc_database_set_options(db, list("some_key" = 123L))
  expect_identical(
    adbc_database_get_option_int(db, "some_key"),
    123L
  )
})

test_that("get/set option can roundtrip double options for database", {
  db <- adbc_database_init(adbc_driver_void())

  expect_error(
    adbc_database_get_option_double(db, "some_key"),
    class = "adbc_status_not_found"
  )

  adbc_database_set_options(db, list("some_key" = 123.4))
  expect_identical(
    adbc_database_get_option_double(db, "some_key"),
    123.4
  )
})


test_that("get/set option can roundtrip string options for connection", {
  db <- adbc_database_init(adbc_driver_void())
  con <- adbc_connection_init(db)

  expect_error(
    adbc_connection_get_option(con, "some_key"),
    class = "adbc_status_not_found"
  )

  adbc_connection_set_options(con, list("some_key" = "some value"))
  expect_identical(
    adbc_connection_get_option(con, "some_key"),
    "some value"
  )
})

test_that("get/set option can roundtrip bytes options for connection", {
  db <- adbc_database_init(adbc_driver_void())
  con <- adbc_connection_init(db)

  expect_error(
    adbc_connection_get_option_bytes(con, "some_key"),
    class = "adbc_status_not_found"
  )

  adbc_connection_set_options(con, list("some_key" = charToRaw("some value")))
  expect_identical(
    adbc_connection_get_option_bytes(con, "some_key"),
    charToRaw("some value")
  )
})

test_that("get/set option can roundtrip int options for connection", {
  db <- adbc_database_init(adbc_driver_void())
  con <- adbc_connection_init(db)

  expect_error(
    adbc_connection_get_option_int(con, "some_key"),
    class = "adbc_status_not_found"
  )

  adbc_connection_set_options(con, list("some_key" = 123L))
  expect_identical(
    adbc_connection_get_option_int(con, "some_key"),
    123L
  )
})

test_that("get/set option can roundtrip double options for connection", {
  db <- adbc_database_init(adbc_driver_void())
  con <- adbc_connection_init(db)

  expect_error(
    adbc_connection_get_option_double(con, "some_key"),
    class = "adbc_status_not_found"
  )

  adbc_connection_set_options(con, list("some_key" = 123.4))
  expect_identical(
    adbc_connection_get_option_double(con, "some_key"),
    123.4
  )
})

test_that("get/set option can roundtrip string options for statement", {
  db <- adbc_database_init(adbc_driver_void())
  con <- adbc_connection_init(db)
  stmt <- adbc_statement_init(con)

  expect_error(
    adbc_statement_get_option(stmt, "some_key"),
    class = "adbc_status_not_found"
  )

  adbc_statement_set_options(stmt, list("some_key" = "some value"))
  expect_identical(
    adbc_statement_get_option(stmt, "some_key"),
    "some value"
  )
})

test_that("get/set option can roundtrip bytes options for statement", {
  db <- adbc_database_init(adbc_driver_void())
  con <- adbc_connection_init(db)
  stmt <- adbc_statement_init(con)

  expect_error(
    adbc_statement_get_option_bytes(stmt, "some_key"),
    class = "adbc_status_not_found"
  )

  adbc_statement_set_options(stmt, list("some_key" = charToRaw("some value")))
  expect_identical(
    adbc_statement_get_option_bytes(stmt, "some_key"),
    charToRaw("some value")
  )
})

test_that("get/set option can roundtrip int options for statement", {
  db <- adbc_database_init(adbc_driver_void())
  con <- adbc_connection_init(db)
  stmt <- adbc_statement_init(con)

  expect_error(
    adbc_statement_get_option_int(stmt, "some_key"),
    class = "adbc_status_not_found"
  )

  adbc_statement_set_options(stmt, list("some_key" = 123L))
  expect_identical(
    adbc_statement_get_option_int(stmt, "some_key"),
    123L
  )
})

test_that("get/set option can roundtrip double options for statement", {
  db <- adbc_database_init(adbc_driver_void())
  con <- adbc_connection_init(db)
  stmt <- adbc_statement_init(con)

  expect_error(
    adbc_statement_get_option_double(stmt, "some_key"),
    class = "adbc_status_not_found"
  )

  adbc_statement_set_options(stmt, list("some_key" = 123.4))
  expect_identical(
    adbc_statement_get_option_double(stmt, "some_key"),
    123.4
  )
})

test_that("void driver errors getting string option of incompatible type", {
  db <- adbc_database_init(adbc_driver_void())
  adbc_database_set_options(db, list("some_key" = "some value"))

  expect_error(
    adbc_database_get_option_int(db, "some_key"),
    class = "adbc_status_not_found"
  )

  expect_error(
    adbc_database_get_option_double(db, "some_key"),
    class = "adbc_status_not_found"
  )
})

test_that("void driver can get string option of compatible type", {
  db <- adbc_database_init(adbc_driver_void())
  adbc_database_set_options(db, list("some_key" = "some value"))

  expect_identical(
    adbc_database_get_option_bytes(db, "some_key"),
    charToRaw("some value")
  )
})

test_that("void driver errors getting bytes option of incorrect type", {
  db <- adbc_database_init(adbc_driver_void())
  adbc_database_set_options(db, list("some_key" = charToRaw("some value")))

  expect_error(
    adbc_database_get_option(db, "some_key"),
    class = "adbc_status_not_found"
  )

  expect_error(
    adbc_database_get_option_int(db, "some_key"),
    class = "adbc_status_not_found"
  )

  expect_error(
    adbc_database_get_option_double(db, "some_key"),
    class = "adbc_status_not_found"
  )
})

test_that("void driver errors getting integer option of incorrect type", {
  db <- adbc_database_init(adbc_driver_void())
  adbc_database_set_options(db, list("some_key" = 123L))

  expect_error(
    adbc_database_get_option(db, "some_key"),
    class = "adbc_status_not_found"
  )

  expect_error(
    adbc_database_get_option_bytes(db, "some_key"),
    class = "adbc_status_not_found"
  )


})

test_that("void driver can get integer option of compatible type", {
  db <- adbc_database_init(adbc_driver_void())
  adbc_database_set_options(db, list("some_key" = 123L))

  expect_identical(
    adbc_database_get_option_double(db, "some_key"),
    123.0
  )
})


test_that("void driver errors getting double option of incorrect type", {
  db <- adbc_database_init(adbc_driver_void())
  adbc_database_set_options(db, list("some_key" = 123.4))

  expect_error(
    adbc_database_get_option(db, "some_key"),
    class = "adbc_status_not_found"
  )

  expect_error(
    adbc_database_get_option_bytes(db, "some_key"),
    class = "adbc_status_not_found"
  )

  expect_error(
    adbc_database_get_option_int(db, "some_key"),
    class = "adbc_status_not_found"
  )
})

test_that("key_value_options works", {
  expect_identical(
    key_value_options(NULL),
    structure(setNames(list(), character()), class = "adbc_options")
  )

  expect_identical(
    key_value_options(c("key" = "value")),
    structure(list("key" = "value"), class = "adbc_options")
  )

  # NULLs dropped
  expect_identical(
    key_value_options(list("key" = "value", "key2" = NULL)),
    structure(list("key" = "value"), class = "adbc_options")
  )

  # Integers stay integers
  expect_identical(
    key_value_options(list("key" = 1L)),
    structure(list("key" = 1L), class = "adbc_options")
  )

  # Doubles stay doubles
  expect_identical(
    key_value_options(list("key" = 1)),
    structure(list("key" = 1), class = "adbc_options")
  )

  # Raw vectors are supported
  expect_identical(
    key_value_options(list("key" = as.raw(c(0x01, 0x05)))),
    structure(list("key" = as.raw(c(0x01, 0x05))), class = "adbc_options")
  )

  # Logical becomes "true" or "false"
  expect_identical(
    key_value_options(list("key" = TRUE, "key2" = FALSE)),
    structure(list("key" = "true", "key2" = "false"), class = "adbc_options")
  )

  # S3 objects converted to strings
  expect_identical(
    key_value_options(list("key" = as.Date("2000-01-01"))),
    structure(list("key" = "2000-01-01"), class = "adbc_options")
  )

  # Can handle many options
  expect_identical(
    key_value_options(setNames(letters, letters)),
    structure(as.list(setNames(letters, letters)), class = "adbc_options")
  )

  # adbc_options() arguments are appended
  expect_identical(
    key_value_options(
      list(
        "key" = "value",
        key_value_options(list("key2" = "value2")),
        "key3" = "value3"
      )
    ),
    structure(
      list("key" = "value", "key2" = "value2", "key3" = "value3"),
      class = "adbc_options"
    )
  )

  # Errors
  expect_error(
    key_value_options(list("value")),
    "must be named"
  )

  expect_error(
    key_value_options(list(key = environment())),
    "Option of type 'environment' (key: 'key') not supported",
    fixed = TRUE
  )

  expect_error(
    key_value_options(setNames(list("value"), "")),
    "must be named"
  )

  expect_error(
    key_value_options(setNames(list("value"), NA_character_)),
    "must be named"
  )
})

# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.


#' Get extended error information from an array stream
#'
#' @param stream A [nanoarrow_array_stream][nanoarrow::as_nanoarrow_array_stream]
#'
#' @return `NULL` if stream was not created by a driver that supports
#'   extended error information or a list whose first element is the
#'   status code and second element is the `adbc_error` object. The
#'   `acbc_error` must not be accessed if `stream` is explicitly released.
#' @export
#'
#' @examples
#' db <- adbc_database_init(adbc_driver_monkey())
#' con <- adbc_connection_init(db)
#' stmt <- adbc_statement_init(con, mtcars)
#' stream <- nanoarrow::nanoarrow_allocate_array_stream()
#' adbc_statement_execute_query(stmt, stream)
#' adbc_error_from_array_stream(stream)
#'
adbc_error_from_array_stream <- function(stream) {
  if (!inherits(stream, "nanoarrow_array_stream") || !adbc_xptr_is_valid(stream)) {
    stop("`stream` must be a valid nanoarrow_array_stream")
  }

  .Call(RAdbcErrorFromArrayStream, stream)
}

adbc_allocate_error <- function(shelter = NULL, use_legacy_error = NULL) {
  if (is.null(use_legacy_error)) {
    use_legacy_error <- getOption("adbcdrivermanager.use_legacy_error", FALSE)
  }

  .Call(RAdbcAllocateError, shelter, use_legacy_error)
}

stop_for_error <- function(status, error) {
  if (!identical(status, 0L)) {
    error <- .Call(RAdbcErrorProxy, error)
    error$status <- status
    error$status_code_message <- .Call(RAdbcStatusCodeMessage, status)
    if (!is.null(error$message)) {
      msg <- paste(error$status_code_message, error$message, sep=": ")
    } else {
      msg <- error$status_code_message
    }

    # Gives an error class like "adbc_status_invalid_state", "adbc_status",
    # "simpleError", ...
    cnd_class <- c(
      paste0("adbc_status_", tolower(gsub("\\s+.*", "", error$status_code_message))),
      "adbc_status"
    )

    # Strips stop_for_error() call from the error output
    cnd <- simpleError(msg, call = sys.call(-1))
    class(cnd) <- union(cnd_class, class(cnd))

    # Store the listified error information
    cnd$error <- error

    stop(cnd)
  }
}

#' @export
print.adbc_error <- function(x, ...) {
  str(x, ...)
}

#' @importFrom utils str
#' @export
str.adbc_error <- function(object, ...) {
  cat("<adbc_error> ")
  str(.Call(RAdbcErrorProxy, object), ...)
  invisible(object)
}

# This is the list()-like interface that allows $ and [[
# to make nice auto-complete when interacting in an IDE

#' @export
length.adbc_error <- function(x, ...) {
  4L
}

#' @export
names.adbc_error <- function(x, ...) {
  c("message", "vendor_code", "sqlstate", "details")
}

#' @export
`[[.adbc_error` <- function(x, i, ...) {
  .Call(RAdbcErrorProxy, x)[[i]]
}

#' @export
`$.adbc_error` <- function(x, i, ...) {
  x[[i]]
}

# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

test_that("adbc_error_from_array_stream() errors for invalid streams", {
  stream <- nanoarrow::nanoarrow_allocate_array_stream()
  expect_error(
    adbc_error_from_array_stream(stream),
    "must be a valid nanoarrow_array_stream"
  )
})

test_that("adbc_error_from_array_stream() returns NULL for unrelated streams", {
  stream <- nanoarrow::basic_array_stream(list(1:5))
  expect_null(adbc_error_from_array_stream(stream))
})

test_that("error allocator works", {
  err <- adbc_allocate_error()
  expect_s3_class(err, "adbc_error")

  expect_output(expect_identical(print(err), err), "adbc_error")
  expect_output(expect_identical(str(err), err), "adbc_error")
  expect_identical(length(err), 4L)
  expect_identical(names(err), c("message", "vendor_code", "sqlstate", "details"))
  expect_null(err$message)
  expect_identical(err$sqlstate, as.raw(c(0x00, 0x00, 0x00, 0x00, 0x00)))
  expect_identical(err$details, setNames(list(), character()))
})

test_that("stop_for_error() gives a custom error class with extra info", {
  had_error <- FALSE
  tryCatch({
    db <- adbc_database_init(adbc_driver_void())
    adbc_database_get_option(db, "this option does not exist")
  }, adbc_status = function(e) {
    had_error <<- TRUE
    expect_s3_class(e, "adbc_status")
    expect_s3_class(e, "adbc_status_not_found")
    expect_identical(e$error$status, 3L)
    expect_identical(
      e$error$detail[["r.driver_test.option_key"]],
      charToRaw("this option does not exist")
    )
  })

  expect_true(had_error)
})

test_that("void driver can report error to ADBC 1.0.0 structs", {
  opts <- options(adbcdrivermanager.use_legacy_error = TRUE)
  on.exit(options(opts))

  had_error <- FALSE
  tryCatch({
    db <- adbc_database_init(adbc_driver_void())
    adbc_database_get_option(db, "this option does not exist")
  }, adbc_status = function(e) {
    had_error <<- TRUE
    expect_s3_class(e, "adbc_status")
    expect_s3_class(e, "adbc_status_not_found")
    expect_identical(e$error$status, 3L)
    expect_identical(e$error$vendor_code, 0L)
    expect_identical(e$error$details, setNames(list(), character()))
  })

  expect_true(had_error)
})

# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

#' Databases
#'
#' @param driver An [adbc_driver()].
#' @param database An [adbc_database][adbc_database_init].
#' @param option A specific option name
#' @param ... Driver-specific options. For the default method, these are
#'   named values that are converted to strings.
#' @param options A named `character()` or `list()` whose values are converted
#'   to strings.
#' @param subclass An extended class for an object so that drivers can specify
#'   finer-grained control over behaviour at the R level.
#'
#' @return An object of class adbc_database
#' @export
#'
#' @examples
#' adbc_database_init(adbc_driver_void())
#'
adbc_database_init <- function(driver, ...) {
  UseMethod("adbc_database_init")
}

#' @export
adbc_database_init.default <- function(driver, ...) {
  adbc_database_init_default(driver, list(...))
}

#' @rdname adbc_database_init
#' @export
adbc_database_init_default <- function(driver, options = NULL, subclass = character()) {
  database <- .Call(
    RAdbcDatabaseNew,
    driver$driver_init_func,
    driver$load_flags
  )

  if (!is.null(driver$name)) {
    adbc_database_set_options(
      database,
      c("driver" = driver$name, "entrypoint" = driver$entrypoint)
    )
  }

  database$driver <- driver

  with_adbc(database, {
    adbc_database_set_options(database, options)

    error <- adbc_allocate_error()
    status <- .Call(RAdbcDatabaseInit, database, error)
    stop_for_error(status, error)
    class(database) <- c(subclass, class(database))

    adbc_xptr_move(database)
  })
}

#' @rdname adbc_database_init
#' @export
adbc_database_release <- function(database) {
  stop_for_nonzero_child_count(database)

  error <- adbc_allocate_error()
  status <- .Call(RAdbcDatabaseRelease, database, error)
  stop_for_error(status, error)
  invisible(database)
}

#' Connections
#'
#' @inheritParams adbc_database_init
#' @param connection An [adbc_connection][adbc_connection_init]
#'
#' @return An object of class 'adbc_connection'
#' @export
#'
#' @examples
#' db <- adbc_database_init(adbc_driver_void())
#' adbc_connection_init(db)
#'
adbc_connection_init <- function(database, ...) {
  UseMethod("adbc_connection_init")
}

#' @export
adbc_connection_init.default <- function(database, ...) {
  adbc_connection_init_default(database, list(...))
}

#' @rdname adbc_connection_init
#' @export
adbc_connection_init_default <- function(database, options = NULL, subclass = character()) {
  connection <- .Call(RAdbcConnectionNew)
  connection$database <- database

  with_adbc(connection, {
    adbc_connection_set_options(connection, options)

    error <- adbc_allocate_error()
    status <- .Call(RAdbcConnectionInit, connection, database, error)
    stop_for_error(status, error)

    class(connection) <- c(subclass, class(connection))

    adbc_xptr_move(connection)
  })
}

#' @rdname adbc_connection_init
#' @export
adbc_connection_release <- function(connection) {
  stop_for_nonzero_child_count(connection)

  if (isTRUE(connection$.release_database)) {
    database <- connection$database
    on.exit(adbc_database_release(database))
  }

  error <- adbc_allocate_error()
  status <- .Call(RAdbcConnectionRelease, connection, error)
  stop_for_error(status, error)
  invisible(connection)
}


#' Connection methods
#'
#' @inheritParams adbc_connection_init
#' @param info_codes A list of metadata codes to fetch, or NULL to fetch all.
#'   Valid values are documented in the adbc.h header.
#' @param depth The level of nesting to display. If 0, display all levels. If 1,
#'   display only catalogs (i.e., catalog_schemas will be null). If 2, display
#'   only catalogs and schemas (i.e., db_schema_tables will be null). If 3,
#'   display only catalogs, schemas, and tables.
#' @param catalog Only show tables in the given catalog. If NULL, do not filter
#'   by catalog. If an empty string, only show tables without a catalog. May be
#'   a search pattern.
#' @param db_schema Only show tables in the given database schema. If NULL, do
#'   not filter by database schema. If an empty string, only show tables without
#'   a database schema. May be a search pattern.
#' @param table_name Constrain an object or statistics query for a specific table.
#'   If NULL, do not filter by name. May be a search pattern.
#' @param table_type Only show tables matching one of the given table types. If
#'   NULL, show tables of any type. Valid table types can be fetched from
#'   GetTableTypes. Terminate the list with a NULL entry.
#' @param column_name Only show columns with the given name. If NULL, do not
#'   filter by name. May be a search pattern.
#' @param serialized_partition The partition descriptor.
#' @param approximate If `FALSE`, request exact values of statistics,
#'   else allow for best-effort, approximate, or cached values. The database
#'   may return approximate values regardless, as indicated in the result.
#'   Requesting exact values may be expensive or unsupported.
#' @param value A string or identifier.
#'
#' @return
#'   - `adbc_connection_get_info()`, `adbc_connection_get_objects()`,
#' `adbc_connection_get_table_types()`, and `adbc_connection_read_partition()`
#' return a [nanoarrow_array_stream][nanoarrow::as_nanoarrow_array_stream()].
#'   - `adbc_connection_get_table_schema()` returns a
#' [nanoarrow_schena][nanoarrow::as_nanoarrow_schema()]
#'   - `adbc_connection_commit()` and `adbc_connection_rollback()` return
#' `connection`, invisibly.
#' @export
#'
#' @examples
#' db <- adbc_database_init(adbc_driver_void())
#' con <- adbc_connection_init(db)
#' # (not implemented by the void driver)
#' try(adbc_connection_get_info(con, 0))
#'
adbc_connection_get_info <- function(connection, info_codes = NULL) {
  error <- adbc_allocate_error()
  out_stream <- nanoarrow::nanoarrow_allocate_array_stream()
  status <- .Call(
    RAdbcConnectionGetInfo,
    connection,
    info_codes,
    out_stream,
    error
  )
  stop_for_error(status, error)

  adbc_child_stream(connection, out_stream)
}

#' @rdname adbc_connection_get_info
#' @export
adbc_connection_get_objects <- function(connection, depth = 0L, catalog = NULL, db_schema = NULL,
                                        table_name = NULL, table_type = NULL, column_name = NULL) {
  error <- adbc_allocate_error()
  out_stream <- nanoarrow::nanoarrow_allocate_array_stream()
  status <- .Call(
    RAdbcConnectionGetObjects,
    connection,
    depth,
    catalog,
    db_schema,
    table_name,
    table_type,
    column_name,
    out_stream,
    error
  )
  stop_for_error(status, error)

  adbc_child_stream(connection, out_stream)
}

#' @rdname adbc_connection_get_info
#' @export
adbc_connection_get_table_schema <- function(connection, catalog, db_schema, table_name) {
  error <- adbc_allocate_error()
  out_schema <- nanoarrow::nanoarrow_allocate_schema()
  status <- .Call(
    RAdbcConnectionGetTableSchema,
    connection,
    catalog,
    db_schema,
    table_name,
    out_schema,
    error
  )
  stop_for_error(status, error)

  out_schema
}

#' @rdname adbc_connection_get_info
#' @export
adbc_connection_get_table_types <- function(connection) {
  error <- adbc_allocate_error()
  out_stream <- nanoarrow::nanoarrow_allocate_array_stream()
  status <- .Call(RAdbcConnectionGetTableTypes, connection, out_stream, error)
  stop_for_error(status, error)

  adbc_child_stream(connection, out_stream)
}

#' @rdname adbc_connection_get_info
#' @export
adbc_connection_read_partition <- function(connection, serialized_partition) {
  error <- adbc_allocate_error()
  out_stream <- nanoarrow::nanoarrow_allocate_array_stream()
  status <- .Call(
    RAdbcConnectionReadPartition,
    connection,
    serialized_partition,
    out_stream,
    error
  )
  stop_for_error(status, error)

  adbc_child_stream(connection, out_stream)
}

#' @rdname adbc_connection_get_info
#' @export
adbc_connection_commit <- function(connection) {
  error <- adbc_allocate_error()
  status <- .Call(RAdbcConnectionCommit, connection, error)
  stop_for_error(status, error)
  invisible(connection)
}


#' @rdname adbc_connection_get_info
#' @export
adbc_connection_rollback <- function(connection) {
  error <- adbc_allocate_error()
  status <- .Call(RAdbcConnectionRollback, connection, error)
  stop_for_error(status, error)
  invisible(connection)
}

#' @rdname adbc_connection_get_info
#' @export
adbc_connection_cancel <- function(connection) {
  error <- adbc_allocate_error()
  status <- .Call(RAdbcConnectionCancel, connection, error)
  stop_for_error(status, error)
  invisible(connection)
}

#' @rdname adbc_connection_get_info
#' @export
adbc_connection_get_statistic_names <- function(connection) {
  error <- adbc_allocate_error()
  out_stream <- nanoarrow::nanoarrow_allocate_array_stream()
  status <- .Call(RAdbcConnectionGetStatisticNames, connection, out_stream, error)
  stop_for_error(status, error)

  adbc_child_stream(connection, out_stream)
}

#' @rdname adbc_connection_get_info
#' @export
adbc_connection_get_statistics <- function(connection, catalog, db_schema,
                                           table_name, approximate = FALSE) {
  error <- adbc_allocate_error()
  out_stream <- nanoarrow::nanoarrow_allocate_array_stream()

  status <- .Call(
    RAdbcConnectionGetStatistics,
    connection,
    catalog,
    db_schema,
    table_name,
    approximate,
    out_stream,
    error
  )
  stop_for_error(status, error)

  adbc_child_stream(connection, out_stream)
}

#' @rdname adbc_connection_get_info
#' @export
adbc_connection_quote_identifier <- function(connection, value, ...) {
  UseMethod("adbc_connection_quote_identifier")
}

#' @rdname adbc_connection_get_info
#' @export
adbc_connection_quote_string <- function(connection, value, ...) {
  UseMethod("adbc_connection_quote_string")
}

#' @export
adbc_connection_quote_identifier.default <- function(connection, value, ...) {
  out <- gsub('"', '""', enc2utf8(value))
  paste0('"', out, '"')
}

#' @export
adbc_connection_quote_string.default <- function(connection, value, ...) {
  out <- gsub("'", "''", enc2utf8(value))
  paste0("'", out, "'")
}

#' Statements
#'
#' @inheritParams adbc_connection_init
#' @param statement An [adbc_statement][adbc_statement_init]
#'
#' @return An object of class 'adbc_statement'
#' @export
#'
#' @examples
#' db <- adbc_database_init(adbc_driver_void())
#' con <- adbc_connection_init(db)
#' adbc_statement_init(con)
#'
adbc_statement_init <- function(connection, ...) {
  UseMethod("adbc_statement_init")
}

#' @export
adbc_statement_init.default <- function(connection, ...) {
  adbc_statement_init_default(connection, list(...))
}

#' @rdname adbc_statement_init
#' @export
adbc_statement_init_default <- function(connection, options = NULL, subclass = character()) {
  statement <- .Call(RAdbcStatementNew, connection)
  statement$connection <- connection

  with_adbc(statement, {
    adbc_statement_set_options(statement, options)
    class(statement) <- c(subclass, class(statement))
    adbc_xptr_move(statement)
  })
}

#' @rdname adbc_statement_init
#' @export
adbc_statement_release <- function(statement) {
  stop_for_nonzero_child_count(statement)

  if (isTRUE(statement$.release_connection)) {
    connection <- statement$connection
    on.exit(adbc_connection_release(connection))
  }

  error <- adbc_allocate_error()
  status <- .Call(RAdbcStatementRelease, statement, error)
  stop_for_error(status, error)
  invisible(statement)
}


#' Statement methods
#'
#' @inheritParams adbc_statement_init
#' @param query An SQL query as a string
#' @param plan A raw vector representation of a serialized Substrait plan.
#' @param values A [nanoarrow_array][nanoarrow::as_nanoarrow_array] or object
#'   that can be coerced to one.
#' @param stream A [nanoarrow_array_stream][nanoarrow::as_nanoarrow_array_stream]
#'   or object that can be coerced to one.
#' @param schema A [nanoarrow_schema][nanoarrow::as_nanoarrow_schema] or object
#'   that can be coerced to one.
#' @param stream_join_parent Use `TRUE` to invalidate `statement` and tie its
#'   lifecycle to `stream`.
#'
#' @return
#'   - `adbc_statement_set_sql_query()`, `adbc_statement_set_substrait_plan()`,
#'     `adbc_statement_prepare()`, `adbc_statement_bind()`,
#'     `adbc_statement_bind_stream()`, and `adbc_statement_execute_query()`
#'     return `statement`, invisibly.
#'   - `adbc_statement_get_parameter_schema()` returns a
#'     [nanoarrow_schema][nanoarrow::as_nanoarrow_schema].
#'
#' @export
#'
#' @examples
#' db <- adbc_database_init(adbc_driver_void())
#' con <- adbc_connection_init(db)
#' stmt <- adbc_statement_init(con)
#' # (not implemented by the void driver)
#' try(adbc_statement_set_sql_query(stmt, "some query"))
#'
adbc_statement_set_sql_query <- function(statement, query) {
  error <- adbc_allocate_error()
  status <- .Call(RAdbcStatementSetSqlQuery, statement, query, error)
  stop_for_error(status, error)
  invisible(statement)
}

#' @rdname adbc_statement_set_sql_query
#' @export
adbc_statement_set_substrait_plan <- function(statement, plan) {
  error <- adbc_allocate_error()
  status <- .Call(RAdbcStatementSetSubstraitPlan, statement, plan, error)
  stop_for_error(status, error)
  invisible(statement)
}

#' @rdname adbc_statement_set_sql_query
#' @export
adbc_statement_prepare <- function(statement) {
  error <- adbc_allocate_error()
  status <- .Call(RAdbcStatementPrepare, statement, error)
  stop_for_error(status, error)
  invisible(statement)
}

#' @rdname adbc_statement_set_sql_query
#' @export
adbc_statement_get_parameter_schema <- function(statement) {
  error <- adbc_allocate_error()
  schema <- nanoarrow::nanoarrow_allocate_schema()
  status <- .Call(RAdbcStatementGetParameterSchema, statement, schema, error)
  stop_for_error(status, error)
  schema
}

#' @rdname adbc_statement_set_sql_query
#' @export
adbc_statement_bind <- function(statement, values, schema = NULL) {
  values <- nanoarrow::as_nanoarrow_array(values, schema = schema)
  schema <- nanoarrow::infer_nanoarrow_schema(values)
  error <- adbc_allocate_error()
  status <- .Call(RAdbcStatementBind, statement, values, schema, error)
  stop_for_error(status, error)
  invisible(statement)
}

#' @rdname adbc_statement_set_sql_query
#' @export
adbc_statement_bind_stream <- function(statement, stream, schema = NULL) {
  stream <- nanoarrow::as_nanoarrow_array_stream(stream, schema = schema)
  error <- adbc_allocate_error()
  status <- .Call(RAdbcStatementBindStream, statement, stream, error)
  stop_for_error(status, error)
  invisible(statement)
}

#' @rdname adbc_statement_set_sql_query
#' @export
adbc_statement_execute_query <- function(statement, stream = NULL,
                                         stream_join_parent = FALSE) {
  error <- adbc_allocate_error()

  if (is.null(stream)) {
    result <- .Call(RAdbcStatementExecuteQuery, statement, NULL, error)
  } else {
    stream_tmp <- nanoarrow::nanoarrow_allocate_array_stream()
    result <- .Call(RAdbcStatementExecuteQuery, statement, stream_tmp, error)
    if (identical(result$status, 0L)) {
      stream_tmp <- adbc_child_stream(
        statement,
        stream_tmp,
        release_parent = stream_join_parent
      )
      nanoarrow::nanoarrow_pointer_export(stream_tmp, stream)
    }
  }

  stop_for_error(result$status, error)
  result$rows_affected
}

#' @rdname adbc_statement_set_sql_query
#' @export
adbc_statement_execute_schema <- function(statement) {
  error <- adbc_allocate_error()
  out_schema <- nanoarrow::nanoarrow_allocate_schema()

  status <- .Call(RAdbcStatementExecuteSchema, statement, out_schema, error)
  stop_for_error(status, error)

  out_schema
}

#' @rdname adbc_statement_set_sql_query
#' @export
adbc_statement_cancel <- function(statement) {
  error <- adbc_allocate_error()
  status <- .Call(RAdbcStatementCancel, statement, error)
  stop_for_error(status, error)
  invisible(statement)
}

# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

test_that("can initialize and release a database", {
  db <- adbc_database_init(adbc_driver_void(), some_key = "some_value")
  expect_s3_class(db, "adbc_database")
  adbc_database_release(db)
  expect_error(adbc_database_release(db), "INVALID_STATE")
})

test_that("can initialize and release a connection", {
  db <- adbc_database_init(adbc_driver_void())
  con <- adbc_connection_init(db, some_key = "some_value")
  expect_s3_class(con, "adbc_connection")
  adbc_connection_release(con)
  expect_error(adbc_connection_release(con), "INVALID_STATE")
})

test_that("connection methods work for the void driver", {
  db <- adbc_database_init(adbc_driver_void())
  con <- adbc_connection_init(db)

  expect_error(
    adbc_connection_get_info(con, integer()),
    "NOT_IMPLEMENTED"
  )

  expect_error(
    adbc_connection_get_info(con, double()),
    "NOT_IMPLEMENTED"
  )

  expect_error(
    adbc_connection_get_info(con, NULL),
    "NOT_IMPLEMENTED"
  )

  # With defaults of NULL/OL
  expect_error(
    adbc_connection_get_objects(con),
    "NOT_IMPLEMENTED"
  )

  # With explicit args
  expect_error(
    adbc_connection_get_objects(
      con, 0,
      "catalog", "db_schema",
      "table_name", "table_type", "column_name"
    ),
    "NOT_IMPLEMENTED"
  )

  expect_error(
    adbc_connection_get_table_schema(
      con,
      "catalog", "db_schema", "table_name"
    ),
    "NOT_IMPLEMENTED"
  )

  expect_error(
    adbc_connection_get_table_types(con),
    "NOT_IMPLEMENTED"
  )

  expect_error(
    adbc_connection_read_partition(con, raw()),
    "NOT_IMPLEMENTED"
  )

  expect_error(
    adbc_connection_commit(con),
    "NOT_IMPLEMENTED"
  )

  expect_error(
    adbc_connection_rollback(con),
    "NOT_IMPLEMENTED"
  )

  expect_error(
    adbc_connection_cancel(con),
    "NOT_IMPLEMENTED"
  )

  expect_error(
    adbc_connection_get_statistic_names(con),
    "NOT_IMPLEMENTED"
  )

  expect_error(
    adbc_connection_get_statistics(con, NULL, NULL, "table name"),
    "NOT_IMPLEMENTED"
  )

  expect_identical(
    adbc_connection_quote_identifier(con, 'some"identifier'),
    '"some""identifier"'
  )

  expect_identical(
    adbc_connection_quote_string(con, "some'value"),
    "'some''value'"
  )
})

test_that("can initialize and release a statement", {
  db <- adbc_database_init(adbc_driver_void())
  con <- adbc_connection_init(db)
  stmt <- adbc_statement_init(con, some_key = "some_value")
  expect_s3_class(stmt, "adbc_statement")
  adbc_statement_release(stmt)
  expect_error(adbc_statement_release(stmt), "INVALID_STATE")
})

test_that("statement methods work for the void driver", {
  db <- adbc_database_init(adbc_driver_void())
  con <- adbc_connection_init(db)
  stmt <- adbc_statement_init(con)

  expect_error(
    adbc_statement_set_sql_query(stmt, "some query"),
    "NOT_IMPLEMENTED"
  )

  expect_error(
    adbc_statement_set_substrait_plan(stmt, charToRaw("some plan")),
    "NOT_IMPLEMENTED"
  )

  expect_error(
    adbc_statement_prepare(stmt),
    "NOT_IMPLEMENTED"
  )

  expect_error(
    adbc_statement_get_parameter_schema(stmt),
    "NOT_IMPLEMENTED"
  )

  struct_array <- nanoarrow::as_nanoarrow_array(data.frame(x = 1:5))
  expect_error(
    adbc_statement_bind(stmt, struct_array),
    "NOT_IMPLEMENTED"
  )

  expect_error(
    adbc_statement_bind_stream(stmt, nanoarrow::nanoarrow_allocate_array_stream()),
    "NOT_IMPLEMENTED"
  )

  expect_error(
    adbc_statement_execute_query(stmt),
    "NOT_IMPLEMENTED"
  )

  expect_error(
    adbc_statement_execute_schema(stmt),
    "NOT_IMPLEMENTED"
  )

  expect_error(
    adbc_statement_cancel(stmt),
    "NOT_IMPLEMENTED"
  )
})

test_that("invalid external pointer inputs generate errors", {
  db <- adbc_database_init(adbc_driver_void())
  con <- adbc_connection_init(db)
  stmt <- adbc_statement_init(con)

  expect_error(
    adbc_database_init(list(driver_init_func = character(), load_flags = 0L)),
    "Expected external pointer with class 'adbc_driver_init_func'"
  )

  expect_error(
    adbc_statement_set_sql_query(con, "some query"),
    "Expected external pointer with class 'adbc_statement'"
  )

  # (makes a NULL xptr)
  stmt2 <- unserialize(serialize(stmt, NULL))
  expect_error(
    adbc_statement_set_sql_query(stmt2, "some query"),
    "Can't convert external pointer to NULL to T*"
  )
})

test_that("invalid integer inputs generate errors", {
  db <- adbc_database_init(adbc_driver_void())
  con <- adbc_connection_init(db)

  expect_error(
    adbc_connection_get_objects(con, depth = "abc"),
    "Expected integer(1) or double(1)",
    fixed = TRUE
  )

  expect_error(
    adbc_connection_get_objects(con, depth = 1:5),
    "Expected integer(1) or double(1)",
    fixed = TRUE
  )

  expect_error(
    adbc_connection_get_objects(con, structure(1L, class = "non-empty")),
    "Can't convert classed object"
  )

  expect_error(
    adbc_connection_get_objects(con, NA_real_),
    "Can't convert non-finite"
  )
})

test_that("invalid int list inputs generate errors", {
  db <- adbc_database_init(adbc_driver_void())
  con <- adbc_connection_init(db)

  expect_error(
    adbc_connection_get_info(con, character()),
    "Expected integer"
  )

  expect_error(
    adbc_connection_get_info(con, structure(integer(), class = "non-empty")),
    "Can't convert classed object"
  )

  expect_error(
    adbc_connection_get_info(con, NA_real_),
    "Can't convert non-finite element"
  )
})

test_that("invalid const char* list inputs generate errors", {
  db <- adbc_database_init(adbc_driver_void())
  con <- adbc_connection_init(db)

  expect_error(
    adbc_connection_get_objects(
      con,
      table_type = integer()
    ),
    "Expected character"
  )

  expect_error(
    adbc_connection_get_objects(
      con,
      table_type = NA_character_
    ),
    "Can't convert NA_character_ element"
  )

  expect_error(
    adbc_connection_get_objects(
      con,
      table_type = structure("abc", class = "non-empty")
    ),
    "Can't convert classed object"
  )
})

test_that("invalid const char* inputs generate errors", {
  db <- adbc_database_init(adbc_driver_void())
  con <- adbc_connection_init(db)
  stmt <- adbc_statement_init(con)

  expect_error(
    adbc_statement_set_sql_query(stmt, NULL),
    "Expected character"
  )

  expect_error(
    adbc_statement_set_sql_query(stmt, structure("abc", class = "non-empty")),
    "Can't convert classed object to const char"
  )

  expect_error(
    adbc_statement_set_sql_query(stmt, NA_character_),
    "Can't convert NA_character_"
  )
})

test_that("invalid bool inputs generate errors", {
  db <- adbc_database_init(adbc_driver_void())
  con <- adbc_connection_init(db)

  expect_error(
    adbc_connection_get_statistics(con, NULL, NULL, "table name", character()),
    "Expected integer(1) or double(1)",
    fixed = TRUE
  )

  expect_error(
    adbc_connection_get_statistics(con, NULL, NULL, "table name", NA),
    "Can't convert NA to bool"
  )

  expect_error(
    adbc_connection_get_statistics(
      con, NULL, NULL, "table name",
      structure(TRUE, class = "non-empty")
    ),
    "Can't convert classed object to bool"
  )
})

# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

test_that("external pointer embedded environment works", {
  db <- adbc_database_init(adbc_driver_void())
  expect_setequal(names(db), c("driver", ".child_count"))
  expect_identical(db$.child_count, 0L)

  db$key <- "value"
  expect_identical(db$key, "value")

  db[["key"]] <- "value2"
  expect_identical(db[["key"]], "value2")
})

test_that("pointer mover leaves behind an invalid external pointer", {
  db <- adbc_database_init(adbc_driver_void())
  expect_true(adbc_xptr_is_valid(db))

  db2 <- adbc_xptr_move(db)
  expect_true(adbc_xptr_is_valid(db2))
  expect_false(adbc_xptr_is_valid(db))

  con <- adbc_connection_init(db2)
  expect_true(adbc_xptr_is_valid(con))

  con2 <- adbc_xptr_move(con)
  expect_true(adbc_xptr_is_valid(con2))
  expect_false(adbc_xptr_is_valid(con))

  stmt <- adbc_statement_init(con2)
  expect_true(adbc_xptr_is_valid(stmt))

  stmt2 <- adbc_xptr_move(stmt)
  expect_true(adbc_xptr_is_valid(stmt2))
  expect_false(adbc_xptr_is_valid(stmt))

  stream <- nanoarrow::basic_array_stream(list(1:5))
  expect_true(adbc_xptr_is_valid(stream))

  stream2 <- adbc_xptr_move(stream)
  expect_true(adbc_xptr_is_valid(stream2))
  expect_false(adbc_xptr_is_valid(stream))
})

test_that("adbc_xptr_is_valid() returns FALSE for null pointer", {
  db <- adbc_database_init(adbc_driver_void())
  con <- adbc_connection_init(db)
  stmt <- adbc_statement_init(con)
  stream <- nanoarrow::basic_array_stream(list(), nanoarrow::na_na())

  # A compact way to set the external pointer to NULL
  db <- unserialize(serialize(db, NULL))
  con <- unserialize(serialize(con, NULL))
  stmt <- unserialize(serialize(stmt, NULL))
  stream <- unserialize(serialize(stream, NULL))

  expect_false(adbc_xptr_is_valid(db))
  expect_false(adbc_xptr_is_valid(con))
  expect_false(adbc_xptr_is_valid(stmt))
  expect_false(adbc_xptr_is_valid(stream))
})

test_that("adbc_xptr_is_valid() errors for non-ADBC objects", {
  expect_error(
    adbc_xptr_is_valid(NULL),
    "must inherit from one of"
  )

  expect_error(
    adbc_xptr_move(NULL),
    "must inherit from one of"
  )
})

# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

test_that("default read/write/execute_adbc() performs the correct calls", {
  # We can't properly test this in the driver manager because we don't
  # have a driver that implements all the methods. Instead, we test this
  # in the sqlite driver and just do a basic log driver snapshot here
  expect_snapshot({
    db <- adbc_database_init(adbc_driver_log())
    try(read_adbc(db, "some sql"))
    try(execute_adbc(db, "some sql"))
    try(write_adbc(mtcars, db, "some_table"))
    adbc_database_release(db)
  })

  expect_snapshot({
    db <- adbc_database_init(adbc_driver_log())
    con <- adbc_connection_init(db)
    try(read_adbc(con, "some sql"))
    try(execute_adbc(con, "some sql"))
    try(write_adbc(mtcars, con, "some_table"))
    adbc_connection_release(con)
    adbc_database_release(db)
  })
})


test_that("with_adbc() and local_adbc() release databases", {
  db <- adbc_database_init(adbc_driver_void())
  expect_identical(with_adbc(db, "value"), "value")
  expect_error(
    adbc_database_release(db),
    "INVALID_STATE"
  )

  db <- adbc_database_init(adbc_driver_void())
  local({
    expect_identical(local_adbc(db), db)
  })
  expect_error(
    adbc_database_release(db),
    "INVALID_STATE"
  )
})

test_that("with_adbc() and local_adbc() release connections", {
  db <- adbc_database_init(adbc_driver_void())
  con <- adbc_connection_init(db)
  expect_identical(with_adbc(con, "value"), "value")
  expect_error(
    adbc_connection_release(con),
    "INVALID_STATE"
  )

  con <- adbc_connection_init(db)
  local({
    expect_identical(local_adbc(con), con)
  })
  expect_error(
    adbc_connection_release(con),
    "INVALID_STATE"
  )
})

test_that("with_adbc() and local_adbc() release statements", {
  db <- adbc_database_init(adbc_driver_void())
  con <- adbc_connection_init(db)
  stmt <- adbc_statement_init(con)
  expect_identical(with_adbc(stmt, "value"), "value")
  expect_error(
    adbc_statement_release(stmt),
    "INVALID_STATE"
  )

  stmt <- adbc_statement_init(con)
  local({
    expect_identical(local_adbc(stmt), stmt)
  })
  expect_error(
    adbc_statement_release(stmt),
    "INVALID_STATE"
  )
})

test_that("with_adbc() and local_adbc() release streams", {
  stream <- nanoarrow::basic_array_stream(list(1:5))
  expect_identical(with_adbc(stream, "value"), "value")
  expect_false(nanoarrow::nanoarrow_pointer_is_valid(stream))

  stream <- nanoarrow::basic_array_stream(list(1:5))
  local({
    expect_identical(local_adbc(stream), stream)
  })
  expect_false(nanoarrow::nanoarrow_pointer_is_valid(stream))
})

test_that("joiners work for databases, connections, and statements", {
  expect_snapshot({
    stmt <- local({
      db <- local_adbc(adbc_database_init(adbc_driver_log()))

      con <- local_adbc(adbc_connection_init(db))
      adbc_connection_join(con, db)
      expect_false(adbc_xptr_is_valid(db))

      stmt <- local_adbc(adbc_statement_init(con))
      adbc_statement_join(stmt, con)
      expect_false(adbc_xptr_is_valid(con))

      adbc_xptr_move(stmt)
    })

    adbc_statement_release(stmt)
  })
})

test_that("joiners work with streams", {
  stream <- local({
    db <- local_adbc(adbc_database_init(adbc_driver_monkey()))

    con <- local_adbc(adbc_connection_init(db))
    adbc_connection_join(con, db)
    expect_false(adbc_xptr_is_valid(db))

    stmt <- local_adbc(adbc_statement_init(con, data.frame(x = 1:5)))
    adbc_statement_join(stmt, con)
    expect_false(adbc_xptr_is_valid(con))

    stream <- local_adbc(nanoarrow::nanoarrow_allocate_array_stream())
    adbc_statement_execute_query(stmt, stream, stream_join_parent = TRUE)
    expect_false(adbc_xptr_is_valid(stmt))

    adbc_xptr_move(stream)
  })

  expect_identical(as.data.frame(stream), data.frame(x = 1:5))
  expect_silent(stream$release())
})

# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

test_that("the monkey driver sees, and the monkey driver does", {
  db <- adbc_database_init(adbc_driver_monkey())
  expect_s3_class(db, "adbc_database_monkey")
  con <- adbc_connection_init(db)
  expect_s3_class(con, "adbc_connection_monkey")

  input <- data.frame(x = 1:10)
  stmt <- adbc_statement_init(con, input)
  expect_s3_class(stmt, "adbc_statement_monkey")
  stream <- nanoarrow::nanoarrow_allocate_array_stream()
  expect_identical(adbc_statement_execute_query(stmt, stream), -1)
  expect_identical(as.data.frame(stream$get_next()), input)
  expect_error(
    adbc_statement_release(stmt),
    class = "adbc_error_child_count_not_zero"
  )
  stream$release()
  adbc_statement_release(stmt)

  stmt <- adbc_statement_init(con, input)
  expect_identical(adbc_statement_execute_query(stmt, NULL), -1)
  adbc_statement_release(stmt)
})

# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

test_that("The log driver logs", {
  expect_snapshot({
    db <- adbc_database_init(adbc_driver_log(), key = "value")
    try(adbc_database_get_option(db, "key"))

    con <- adbc_connection_init(db, key = "value")
    try(adbc_connection_get_option(con, "key"))
    try(adbc_connection_commit(con))
    try(adbc_connection_get_info(con))
    try(adbc_connection_get_objects(con))
    try(adbc_connection_get_table_schema(con, NULL, NULL, "table_name"))
    try(adbc_connection_get_table_types(con))
    try(adbc_connection_read_partition(con, raw()))
    try(adbc_connection_rollback(con))
    try(adbc_connection_cancel(con))
    try(adbc_connection_get_statistics(con, NULL, NULL, "table_name"))
    try(adbc_connection_get_statistic_names(con))

    stmt <- adbc_statement_init(con, key = "value")
    try(adbc_statement_get_option(stmt, "key"))

    try(adbc_statement_execute_query(stmt))
    try(adbc_statement_execute_schema(stmt))
    try(adbc_statement_prepare(stmt))
    try(adbc_statement_set_sql_query(stmt, ""))
    try(adbc_statement_set_substrait_plan(stmt, raw()))
    try(adbc_statement_bind(stmt, data.frame()))
    try(adbc_statement_bind_stream(stmt, data.frame()))
    try(adbc_statement_cancel(stmt))

    adbc_statement_release(stmt)
    adbc_connection_release(con)
    adbc_database_release(db)
  })
})

# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

#' Log calls to another driver
#'
#' Useful for debugging or ensuring that certain calls occur during
#' initialization and/or cleanup. The current logging output should not
#' be considered stable and may change in future releases.
#'
#'
#' @return An object of class 'adbc_driver_log'
#' @export
#'
#' @examples
#' drv <- adbc_driver_log()
#' db <- adbc_database_init(drv, key = "value")
#' con <- adbc_connection_init(db, key = "value")
#' stmt <- adbc_statement_init(con, key = "value")
#' try(adbc_statement_execute_query(stmt))
#' adbc_statement_release(stmt)
#' adbc_connection_release(con)
#' adbc_database_release(db)
#'
adbc_driver_log <- function() {
  if (is.null(internal_driver_env$log)) {
    internal_driver_env$log <- adbc_driver(
      .Call(RAdbcLogDriverInitFunc),
      subclass = "adbc_driver_log"
    )
  }

  internal_driver_env$log
}

#' @export
adbc_database_init.adbc_driver_log <- function(driver, ...) {
  adbc_database_init_default(
    driver,
    options = list(...),
    subclass = "adbc_database_log"
  )
}

#' @export
adbc_connection_init.adbc_database_log <- function(database, ...) {
  adbc_connection_init_default(
    database,
    options = list(...),
    subclass = "adbc_connection_log"
  )
}

#' @export
adbc_statement_init.adbc_connection_log <- function(connection, ...) {
  adbc_statement_init_default(
    connection,
    options = list(...),
    subclass = "adbc_statement_log"
  )
}

# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

new_env <- function() {
  env <- new.env(parent = emptyenv())
  # A previous version of this just did env$.child_count <- 0L,
  # which, perhaps because of compilation, results in env$.child_count
  # referring to the exact same SEXP for every ADBC object! Use vector()
  # to ensure a fresh allocation.
  env$.child_count <- vector("integer", length = 1L)
  env
}

xptr_env <- function(xptr) {
  .Call(RAdbcXptrEnv, xptr)
}

xptr_set_protected <- function(xptr, prot) {
  .Call(RAdbcXptrSetProtected, xptr, prot)
}

#' @export
length.adbc_xptr <- function(x) {
  length(xptr_env(x))
}

#' @export
names.adbc_xptr <- function(x) {
  names(xptr_env(x))
}


#' @export
`[[.adbc_xptr` <- function(x, i) {
  xptr_env(x)[[i]]
}

#' @export
`[[<-.adbc_xptr` <- function(x, i, value) {
  env <- xptr_env(x)
  env[[i]] <- value
  invisible(x)
}

#' @export
`$.adbc_xptr` <- function(x, name) {
  xptr_env(x)[[name]]
}

#' @export
`$<-.adbc_xptr` <- function(x, name, value) {
  env <- xptr_env(x)
  env[[name]] <- value
  invisible(x)
}

#' @export
print.adbc_xptr <- function(x, ...) {
  str(x, ...)
}

#' @export
format.adbc_xptr <- function(x, ...) {
  sprintf(
    "<%s at %s> ",
    class(x)[1],
    nanoarrow::nanoarrow_pointer_addr_pretty(x)
  )
}

#' @export
str.adbc_xptr <- function(object, ...) {
  cat(sprintf("%s\n", format(object)))
  env_proxy <- as.list(xptr_env(object))
  str(env_proxy, ...)
  invisible(object)
}

stop_for_nonzero_child_count <- function(obj) {
  child_count <- obj$.child_count
  if (!identical(child_count, 0L)) {
    msg <- sprintf(
      "<%s> has %d unreleased child object%s",
      paste(class(obj), collapse = "/"),
      child_count,
      if (child_count != 1) "s" else ""
    )
    cnd <- simpleError(msg, call = sys.call(-1))
    class(cnd) <- union("adbc_error_child_count_not_zero", class(cnd))
    stop(cnd)
  }
}

#' Low-level pointer details
#'
#' - `adbc_xptr_move()` allocates a fresh R object and moves all values pointed
#'   to by `x` into it. The original R object is invalidated by zeroing its
#'   content. This is useful when returning from a function where
#'   [lifecycle helpers][with_adbc] were used to manage the original
#'   object.
#' - `adbc_xptr_is_valid()` provides a means by which to test for an invalidated
#'   pointer.
#'
#' @param x An 'adbc_database', 'adbc_connection', 'adbc_statement', or
#'   'nanoarrow_array_stream'
#' @param check_child_count Ensures that `x` has a zero child count before
#'   performing the move. This should almost always be `TRUE`.
#'
#' @return
#' - `adbc_xptr_move()`: A freshly-allocated R object identical to `x`
#' - `adbc_xptr_is_valid()`: Returns FALSE if the ADBC object pointed to by `x`
#'   has been invalidated.
#' @export
#'
#' @examples
#' db <- adbc_database_init(adbc_driver_void())
#' adbc_xptr_is_valid(db)
#' db_new <- adbc_xptr_move(db)
#' adbc_xptr_is_valid(db)
#' adbc_xptr_is_valid(db_new)
#'
adbc_xptr_move <- function(x, check_child_count = TRUE) {
  if (check_child_count && (".child_count" %in% names(x))) {
    stop_for_nonzero_child_count(x)
  }

  if (inherits(x, "adbc_database")) {
    .Call(RAdbcMoveDatabase, x)
  } else if (inherits(x, "adbc_connection")) {
    .Call(RAdbcMoveConnection, x)
  } else if (inherits(x, "adbc_statement")) {
    .Call(RAdbcMoveStatement, x)
  } else if (inherits(x, "nanoarrow_array_stream")) {
    stream <- nanoarrow::nanoarrow_allocate_array_stream()
    nanoarrow::nanoarrow_pointer_move(x, stream)
    stream
  } else {
    assert_adbc(x)
  }
}

#' @rdname adbc_xptr_move
#' @export
adbc_xptr_is_valid <- function(x) {
  if (inherits(x, "adbc_database")) {
    .Call(RAdbcDatabaseValid, x)
  } else if (inherits(x, "adbc_connection")) {
    .Call(RAdbcConnectionValid, x)
  } else if (inherits(x, "adbc_statement")) {
    .Call(RAdbcStatementValid, x)
  } else if (inherits(x, "nanoarrow_array_stream")) {
    nanoarrow::nanoarrow_pointer_is_valid(x)
  } else {
    assert_adbc(x)
  }
}

# Usually we want errors for an attempt at double release; however,
# the helpers we want to be compatible with adbc_xptr_move() which sets the
# managed pointer to NULL.
adbc_release_non_null <- function(x) {
  if (!adbc_xptr_is_valid(x)) {
    return()
  }

  if (inherits(x, "adbc_database")) {
    adbc_database_release(x)
  } else if (inherits(x, "adbc_connection")) {
    adbc_connection_release(x)
  } else if (inherits(x, "adbc_statement")) {
    adbc_statement_release(x)
  } else if (inherits(x, "nanoarrow_array_stream")) {
    nanoarrow::nanoarrow_pointer_release(x)
  } else {
    assert_adbc(x)
  }
}

adbc_classes <- c(
  "adbc_database", "adbc_connection", "adbc_statement",
  "nanoarrow_array_stream"
)

assert_adbc <- function(x, what = adbc_classes) {
  if (inherits(x, what)) {
    return(invisible(x))
  }

  stop(
    sprintf(
      "`x` must inherit from one of: %s",
      paste0("'", what, "'", collapse = ", ")
    ),
    call. = sys.call(-1)
  )
}

# Locate the shared library for tests that load drivers
adbcdrivermanager_shared <- function() {
  # The location of the shared library when installed
  shared <- system.file(
    "libs",
    .Platform$r_arch,
    paste0("adbcdrivermanager", .Platform$dynlib.ext),
    package = "adbcdrivermanager"
  )

  if (!identical(shared, "")) {
    return(shared)
  }

  # The location of the shared library in interactive development
  system.file(
    "src",
    paste0("adbcdrivermanager", .Platform$dynlib.ext),
    package = "adbcdrivermanager"
  )
}

current_arch <- function() {
  .Call(RAdbcCurrentArch)
}

# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

#' Read, write, and execute on ADBC connections
#'
#' These are convenience methods useful for testing connections. Note that
#' S3 dispatch is always on `db_or_con` (i.e., drivers may provide their own
#' implementations).
#'
#' @param db_or_con An adbc_database or adbc_connection. If a database, a
#'   connection will be opened. For `read_adbc()`, this connection will
#'   be closed when the resulting stream has been released.
#' @param tbl A data.frame, [nanoarrow_array][nanoarrow::as_nanoarrow_array],
#'   or  [nanoarrow_array_stream][nanoarrow::as_nanoarrow_array_stream].
#' @param target_table A target table name to which `tbl` should be written.
#' @param mode One of `"create"`, `"append"`, `"replace"`, `"create_append"` (error if the schema
#'   is not compatible or append otherwise), or `"default"` (use the `adbc.ingest.mode`
#'   argument of [adbc_statement_init()]). The default is `"default"`.
#' @param query An SQL query
#' @param bind A data.frame, nanoarrow_array, or nanoarrow_array_stream of
#'   bind parameters or NULL to skip the bind/prepare step.
#' @param temporary Use TRUE to create a table as a temporary table.
#' @param catalog_name If not `NULL`, the catalog to create/locate the table in.
#' @param db_schema_name If not `NULL`, the schema to create/locate the table in.
#' @param ... Passed to S3 methods.
#'
#' @return
#'   - `read_adbc()`: A [nanoarrow_array_stream][nanoarrow::as_nanoarrow_array_stream]
#'   - `execute_adbc()`: `db_or_con`, invisibly.
#'   - `write_adbc()`: `tbl`, invisibly.
#' @export
#'
#' @examples
#' # On a database, connections are opened and closed
#' db <- adbc_database_init(adbc_driver_log())
#' try(read_adbc(db, "some sql"))
#' try(execute_adbc(db, "some sql"))
#' try(write_adbc(mtcars, db, "some_table"))
#'
#' # Also works on a connection
#' con <- adbc_connection_init(db)
#' try(read_adbc(con, "some sql"))
#' try(execute_adbc(con, "some sql"))
#' try(write_adbc(mtcars, con, "some_table"))
#'
read_adbc <- function(db_or_con, query, ..., bind = NULL) {
  UseMethod("read_adbc")
}

#' @rdname read_adbc
#' @export
execute_adbc <- function(db_or_con, query, ..., bind = NULL) {
  UseMethod("execute_adbc")
}

#' @rdname read_adbc
#' @export
write_adbc <- function(tbl, db_or_con, target_table, ...,
                       mode = c("default", "create", "append", "replace", "create_append"),
                       temporary = FALSE,
                       catalog_name = NULL,
                       db_schema_name = NULL) {
  UseMethod("write_adbc", db_or_con)
}


#' @export
read_adbc.default <- function(db_or_con, query, ..., bind = NULL) {
  stream <- nanoarrow::nanoarrow_allocate_array_stream()
  execute_adbc(db_or_con, query, ..., bind = bind, stream = stream)
  stream
}

#' @export
execute_adbc.default <- function(db_or_con, query, ..., bind = NULL, stream = NULL) {
  assert_adbc(db_or_con, c("adbc_database", "adbc_connection"))

  if (inherits(db_or_con, "adbc_database")) {
    con <- adbc_connection_init(db_or_con)
    on.exit(adbc_connection_release(con))

    stmt <- adbc_statement_init(con)
    adbc_statement_join(stmt, con)

  } else {
    con <- db_or_con
    stmt <- adbc_statement_init(con)
  }

  on.exit(adbc_release_non_null(stmt))
  adbc_statement_set_sql_query(stmt, query)

  if (!is.null(bind)) {
    adbc_statement_bind_stream(stmt, bind)
    adbc_statement_prepare(stmt)
  }

  adbc_statement_execute_query(stmt, stream, stream_join_parent = TRUE)

  invisible(db_or_con)
}

#' @export
write_adbc.default <- function(tbl, db_or_con, target_table, ...,
                               mode = c("default", "create", "append", "replace", "create_append"),
                               temporary = FALSE,
                               catalog_name = NULL,
                               db_schema_name = NULL) {
  assert_adbc(db_or_con, c("adbc_database", "adbc_connection"))
  mode <- match.arg(mode)

  if (inherits(db_or_con, "adbc_database")) {
    con <- adbc_connection_init(db_or_con)
    on.exit(adbc_connection_release(con))
  } else {
    con <- db_or_con
  }

  stmt <- adbc_statement_init(
    con,
    adbc.ingest.target_table = target_table,
    adbc.ingest.target_catalog = catalog_name,
    adbc.ingest.target_db_schema = db_schema_name,
    adbc.ingest.mode = if (!identical(mode, "default")) paste0("adbc.ingest.mode.", mode),
    adbc.ingest.temporary = if (temporary) "true"
  )
  on.exit(adbc_statement_release(stmt), add = TRUE, after = FALSE)

  adbc_statement_bind_stream(stmt, tbl)
  adbc_statement_execute_query(stmt)
  invisible(tbl)
}


#' Cleanup helpers
#'
#' Managing the lifecycle of databases, connections, and statements can
#' be complex and error-prone. The R objects that wrap the underlying ADBC
#' pointers will perform cleanup in the correct order if you rely on garbage
#' collection (i.e., do nothing and let the objects go out of scope); however
#' it is good practice to explicitly clean up these objects. These helpers
#' are designed to make explicit and predictable cleanup easy to accomplish.
#'
#' Note that you can use [adbc_connection_join()] and [adbc_statement_join()]
#' to tie the lifecycle of the parent object to that of the child object.
#' These functions mark any previous references to the parent object as
#' released so you can still use local and with helpers to manage the parent
#' object before it is joined. Use `stream_join_parent = TRUE` in
#' [adbc_statement_execute_query()] to tie the lifecycle of a statement to
#' the output stream.
#'
#' @param x An ADBC database, ADBC connection, ADBC statement, or
#'   nanoarrow_array_stream returned from calls to an ADBC function.
#' @param code Code to execute before cleaning up the input.
#' @param .local_envir The execution environment whose scope should be tied
#'   to the input.
#'
#' @return
#'   - `with_adbc()` returns the result of `code`
#'   - `local_adbc()` returns the input, invisibly.
#' @export
#'
#' @examples
#' # Using with_adbc():
#' with_adbc(db <- adbc_database_init(adbc_driver_void()), {
#'   with_adbc(con <- adbc_connection_init(db), {
#'     with_adbc(stmt <- adbc_statement_init(con), {
#'       # adbc_statement_set_sql_query(stmt, "SELECT * FROM foofy")
#'       # adbc_statement_execute_query(stmt)
#'       "some result"
#'     })
#'   })
#' })
#'
#' # Using local_adbc_*() (works best within a function, test, or local())
#' local({
#'   db <- local_adbc(adbc_database_init(adbc_driver_void()))
#'   con <- local_adbc(adbc_connection_init(db))
#'   stmt <- local_adbc(adbc_statement_init(con))
#'   # adbc_statement_set_sql_query(stmt, "SELECT * FROM foofy")
#'   # adbc_statement_execute_query(stmt)
#'   "some result"
#' })
#'
with_adbc <- function(x, code) {
  assert_adbc(x)

  on.exit(adbc_release_non_null(x))
  force(code)
}

#' @rdname with_adbc
#' @export
local_adbc <- function(x, .local_envir = parent.frame()) {
  assert_adbc(x)

  withr::defer(adbc_release_non_null(x), envir = .local_envir)
  invisible(x)
}

#' Join the lifecycle of a unique parent to its child
#'
#' It is occasionally useful to return a connection, statement, or stream
#' from a function that was created from a unique parent. These helpers
#' tie the lifecycle of a unique parent object to its child such that the
#' parent object is released predictably and immediately after the child.
#' These functions will invalidate all references to the previous R object.
#'
#' @param database A database created with [adbc_database_init()]
#' @param connection A connection created with [adbc_connection_init()]
#' @param statement A statement created with [adbc_statement_init()]
#'
#' @return The input, invisibly.
#' @export
#'
#' @examples
#' # Use local_adbc to ensure prompt cleanup on error;
#' # use join functions to return a single object that manages
#' # the lifecycle of all three.
#' stmt <- local({
#'   db <- local_adbc(adbc_database_init(adbc_driver_log()))
#'
#'   con <- local_adbc(adbc_connection_init(db))
#'   adbc_connection_join(con, db)
#'
#'   stmt <- local_adbc(adbc_statement_init(con))
#'   adbc_statement_join(stmt, con)
#'
#'   adbc_xptr_move(stmt)
#' })
#'
#' # Everything is released immediately when the last object is released
#' adbc_statement_release(stmt)
#'
adbc_connection_join <- function(connection, database) {
  assert_adbc(connection, "adbc_connection")

  stopifnot(
    identical(database, connection$database),
    identical(database$.child_count, 1L)
  )

  connection$.release_database <- TRUE
  connection$database <- adbc_xptr_move(database, check_child_count = FALSE)
  xptr_set_protected(connection, connection$database)
  invisible(connection)
}

#' @rdname adbc_connection_join
#' @export
adbc_statement_join <- function(statement, connection) {
  assert_adbc(statement, "adbc_statement")

  stopifnot(
    identical(connection, statement$connection),
    identical(connection$.child_count, 1L)
  )

  statement$.release_connection <- TRUE
  statement$connection <- adbc_xptr_move(connection, check_child_count = FALSE)
  xptr_set_protected(statement, statement$connection)
  invisible(statement)
}

adbc_child_stream <- function(parent, stream, release_parent = FALSE) {
  assert_adbc(parent)

  # This finalizer will run immediately on release (if released explicitly
  # on the main R thread) or on garbage collection otherwise.
  self_contained_finalizer <- function() {
    try({
      parent$.child_count <- parent$.child_count - 1L
      if (release_parent) {
        adbc_release_non_null(parent)
      }
    })
  }

  # Make sure we don't keep any variables around that aren't needed
  # for the finalizer and make sure we do keep around a strong reference
  # to parent.
  self_contained_finalizer_env <- as.environment(
    list(
      parent = if (release_parent) adbc_xptr_move(parent) else parent,
      release_parent = release_parent
    )
  )
  parent.env(self_contained_finalizer_env) <- asNamespace("adbcdrivermanager")
  environment(self_contained_finalizer) <- self_contained_finalizer_env

  # Set the finalizer using nanoarrow's method for this
  stream_out <- nanoarrow::array_stream_set_finalizer(
    stream,
    self_contained_finalizer
  )

  # Once we're sure this will succeed, increment the parent child count
  # Use whatever version is in the finalizer env (we might have moved parent)
  self_contained_finalizer_env$parent$.child_count <-
    self_contained_finalizer_env$parent$.child_count + 1L
  stream_out
}

# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

#' Monkey see, monkey do!
#'
#' A driver whose query results are set in advance.
#'
#' @return An object of class 'adbc_driver_monkey'
#' @export
#'
#' @examples
#' db <- adbc_database_init(adbc_driver_monkey())
#' con <- adbc_connection_init(db)
#' stmt <- adbc_statement_init(con, mtcars)
#' stream <- nanoarrow::nanoarrow_allocate_array_stream()
#' adbc_statement_execute_query(stmt, stream)
#' as.data.frame(stream$get_next())
#'
adbc_driver_monkey <- function() {
  if (is.null(internal_driver_env$monkey)) {
    internal_driver_env$monkey <- adbc_driver(
      .Call(RAdbcMonkeyDriverInitFunc),
      subclass = "adbc_driver_monkey"
    )
  }

  internal_driver_env$monkey
}

#' @export
adbc_database_init.adbc_driver_monkey <- function(driver, ...) {
  adbc_database_init_default(driver, subclass = "adbc_database_monkey")
}

#' @export
adbc_connection_init.adbc_database_monkey <- function(database, ...) {
  adbc_connection_init_default(database, subclass = "adbc_connection_monkey")
}

#' @export
adbc_statement_init.adbc_connection_monkey <- function(connection, stream = NULL, ...) {
  stmt <- adbc_statement_init_default(connection, subclass = "adbc_statement_monkey")

  if (!is.null(stream)) {
    adbc_statement_bind_stream(stmt, stream)
  }

  stmt
}

# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

#' Create ADBC drivers
#'
#' Creates the R object representation of an ADBC driver, which consists of a
#' name and an initializer function with an optional subclass to control
#' finer-grained behaviour at the R level.
#'
#' @param x,entrypoint An ADBC driver may be defined either as an init function
#'   or as an identifier with an entrypoint name. A driver init func
#'   must be an external pointer to a DL_FUNC with the type
#'   `AdbcDriverInitFunc` specified in the adbc.h header.
#' @param load_flags Integer flags generated by [adbc_load_flags()]
#' @param ... Further key/value parameters to store with the (R-level) driver
#'   object.
#' @param subclass An optional subclass for finer-grained control of
#'   behaviour at the R level.
#'
#' @return An object of class 'adbc_driver'
#' @export
#'
#' @examples
#' adbc_driver_void()
#'
adbc_driver_void <- function() {
  if (is.null(internal_driver_env$void)) {
    internal_driver_env$void <- adbc_driver(
      .Call(RAdbcVoidDriverInitFunc),
      subclass = "adbc_driver_void"
    )
  }

  internal_driver_env$void
}

#' @rdname adbc_driver_void
#' @export
adbc_driver <- function(x, entrypoint = NULL, ...,
                        load_flags = adbc_load_flags(),
                        subclass = character()) {
  error <- adbc_allocate_error()
  driver_alloc <- .Call(RAdbcAllocateDriver)
  driver <- driver_alloc$driver
  version <- driver_alloc$version

  # Attempt to load the driver to ensure it works
  status <- adbc_driver_load(x, entrypoint, version, driver, error, load_flags)
  stop_for_error(status, error)

  # Keep track of the version/mechanism used to successfully load a driver
  driver$version <- version
  driver$load_flags <- load_flags
  if (inherits(x, "adbc_driver_init_func")) {
    driver$driver_init_func <- x
  } else {
    driver$name <- x
    driver$entrypoint <- entrypoint
  }

  # Add extra properties
  args <- list(...)
  for (i in seq_along(args)) {
    driver[[names(args)[i]]] <- args[[i]]
  }

  class(driver) <- c(subclass, class(driver))
  driver
}

#' Low-level driver loader
#'
#' Most users should use [adbc_driver()]; however, this function may be used
#' to allow other libraries (e.g., GDAL) to access the driver loader.
#'
#' @inheritParams adbc_driver_void
#' @param driver An external pointer to an `AdbcDriver`
#' @param version The version number corresponding to the `driver` supplied
#' @param error An external pointer to an `AdbcError` or NULL
#' @param additional_search_path_list A path list of additional locations to search for driver manifests
#'
#' @return An integer ADBC status code
#' @export
#'
adbc_driver_load <- function(x, entrypoint, version, driver, error,
                             load_flags = adbc_load_flags(), additional_search_path_list = NULL) {
  if (inherits(x, "adbc_driver_init_func")) {
    .Call(RAdbcLoadDriverFromInitFunc, x, version, driver, error)
  } else {
    .Call(RAdbcLoadDriver, x, entrypoint, version, load_flags, additional_search_path_list, driver, error)
  }
}

#' Driver search/load options
#'
#' Options that indicate where to look for driver manifests. Manifests
#' (.toml files) can be installed at the system level, the user level,
#' in location(s) specified by the ADBC_DRIVER_PATH environment variable,
#' and/or in a conda environment. See the ADBC documentation for details
#' regarding the locations of the user and system paths on various platforms.
#'
#' @param search_env Search for manifest files in the directories specified in
#'   the ADBC_DRIVER_PATH environment variable and (when installed with conda)
#'   in the conda environment.
#' @param search_user Search for manifest files in the designated directory
#'   for user ADBC driver installs.
#' @param search_system Search for manifest files in the designtaed directory
#'   for system ADBC driver installs.
#' @param allow_relative_paths Allow shared objects to be specified relative to
#'   the current working directory.
#'
#' @return An integer flag value for use in `adbc_driver()`
#'
adbc_load_flags <- function(search_env = TRUE, search_user = TRUE,
                            search_system = TRUE, allow_relative_paths = TRUE) {
  load_flags <- 0L

  # define ADBC_LOAD_FLAG_SEARCH_ENV 1
  # define ADBC_LOAD_FLAG_SEARCH_USER 2
  # define ADBC_LOAD_FLAG_SEARCH_SYSTEM 4
  # define ADBC_LOAD_FLAG_ALLOW_RELATIVE_PATHS 8
  if (search_env) {
    load_flags <- bitwOr(load_flags, 1L)
  }

  if (search_user) {
    load_flags <- bitwOr(load_flags, 2L)
  }

  if (search_system) {
    load_flags <- bitwOr(load_flags, 4L)
  }

  if (allow_relative_paths) {
    load_flags <- bitwOr(load_flags, 8L)
  }

  load_flags
}

internal_driver_env <- new.env(parent = emptyenv())

#' @export
print.adbc_driver <- function(x, ...) {
  str(x, ...)
}

#' @importFrom utils str
#' @export
str.adbc_driver <- function(object, ...) {
  cat(sprintf("<%s> ", class(object)[1]))

  fields <- names(object)
  object_lst <- Map("[[", list(object), fields)
  names(object_lst) <- fields

  str(object_lst, ...)
  invisible(object)
}

# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

test_that("void driver init function works", {
  expect_s3_class(adbc_driver_void(), "adbc_driver")
  expect_s3_class(adbc_driver_void()$driver_init_func, "adbc_driver_init_func")
})

test_that("drivers can be loaded by name/entrypoint", {
  shared <- adbcdrivermanager_shared()
  driver <- adbc_driver(shared, "AdbcTestVoidDriverInit")
  expect_s3_class(driver, "adbc_driver")
  expect_identical(driver$name, shared)
})

test_that("drivers are loaded using load_flags", {
  # Linux, at least in GitHub CI, doesn't allow loading by relative path
  # with either flag configuration.
  skip_on_os("linux")

  shared <- adbcdrivermanager_shared()

  withr::with_dir(dirname(shared), {
    expect_s3_class(
      adbc_driver(basename(shared), "AdbcTestVoidDriverInit"),
      "adbc_driver"
    )

    # Check that load_flags are used when test-loading the driver
    expect_error(
      adbc_driver(
        basename(shared),
        load_flags = adbc_load_flags(allow_relative_paths = FALSE)
      ),
      "Driver path is relative and relative paths are not allowed"
    )

    # Check that load_flags are passed to the database
    drv <- adbc_driver(basename(shared), "AdbcTestVoidDriverInit")
    drv$load_flags <- adbc_load_flags(allow_relative_paths = FALSE)
    expect_error(
      adbc_database_init(drv),
      "Driver path is relative and relative paths are not allowed"
    )
  })
})

test_that("drivers can be loaded by manifest path", {
  toml_content <- sprintf("
manifest_version = 1

name = 'Void Driver'

[ADBC]
version = 'v1.1.0'

[Driver]
entrypoint = 'AdbcTestVoidDriverInit'

[Driver.shared]
%s = '%s'

  ", current_arch(), adbcdrivermanager_shared())

  td <- tempfile()
  toml_path <- file.path(td, "void.toml")
  on.exit(unlink(td))
  dir.create(td)
  writeLines(toml_content, toml_path)

  drv <- adbc_driver(toml_path)
  expect_s3_class(adbc_database_init(drv), "adbc_database")
})

# The log driver logs

    Code
      db <- adbc_database_init(adbc_driver_log(), key = "value")
    Output
      LogDatabaseNew()
      LogDatabaseSetOption()
      LogDatabaseInit()
    Code
      try(adbc_database_get_option(db, "key"))
    Output
      LogDatabaseGetOption()
      Error in adbc_database_get_option(db, "key") : NOT_FOUND: Unknown option
    Code
      con <- adbc_connection_init(db, key = "value")
    Output
      LogConnectionNew()
      LogConnectionSetOption()
      LogConnectionInit()
    Code
      try(adbc_connection_get_option(con, "key"))
    Output
      Error in adbc_connection_get_option(con, "key") : 
        NOT_FOUND: Unknown option
    Code
      try(adbc_connection_commit(con))
    Output
      LogConnectionCommit()
      Error in adbc_connection_commit(con) : NOT_IMPLEMENTED
    Code
      try(adbc_connection_get_info(con))
    Output
      LogConnectionGetInfo()
      Error in adbc_connection_get_info(con) : NOT_IMPLEMENTED
    Code
      try(adbc_connection_get_objects(con))
    Output
      LogConnectionGetObjects()
      Error in adbc_connection_get_objects(con) : NOT_IMPLEMENTED
    Code
      try(adbc_connection_get_table_schema(con, NULL, NULL, "table_name"))
    Output
      LogConnectionGetTableSchema()
      Error in adbc_connection_get_table_schema(con, NULL, NULL, "table_name") : 
        NOT_IMPLEMENTED
    Code
      try(adbc_connection_get_table_types(con))
    Output
      LogConnectionGetTableTypes()
      Error in adbc_connection_get_table_types(con) : NOT_IMPLEMENTED
    Code
      try(adbc_connection_read_partition(con, raw()))
    Output
      LogConnectionReadPartition()
      Error in adbc_connection_read_partition(con, raw()) : NOT_IMPLEMENTED
    Code
      try(adbc_connection_rollback(con))
    Output
      LogConnectionRollback()
      Error in adbc_connection_rollback(con) : NOT_IMPLEMENTED
    Code
      try(adbc_connection_cancel(con))
    Output
      LogConnectionCancel()
      Error in adbc_connection_cancel(con) : NOT_IMPLEMENTED
    Code
      try(adbc_connection_get_statistics(con, NULL, NULL, "table_name"))
    Output
      LogConnectionGetStatistics()
      Error in adbc_connection_get_statistics(con, NULL, NULL, "table_name") : 
        NOT_IMPLEMENTED
    Code
      try(adbc_connection_get_statistic_names(con))
    Output
      LogConnectionGetStatisticNames()
      Error in adbc_connection_get_statistic_names(con) : NOT_IMPLEMENTED
    Code
      stmt <- adbc_statement_init(con, key = "value")
    Output
      LogStatementNew()
      LogStatementSetOption()
    Code
      try(adbc_statement_get_option(stmt, "key"))
    Output
      Error in adbc_statement_get_option(stmt, "key") : 
        NOT_FOUND: Unknown option
    Code
      try(adbc_statement_execute_query(stmt))
    Output
      LogStatementExecuteQuery()
      Error in adbc_statement_execute_query(stmt) : NOT_IMPLEMENTED
    Code
      try(adbc_statement_execute_schema(stmt))
    Output
      LogStatementExecuteSchema()
      Error in adbc_statement_execute_schema(stmt) : NOT_IMPLEMENTED
    Code
      try(adbc_statement_prepare(stmt))
    Output
      LogStatementPrepare()
      Error in adbc_statement_prepare(stmt) : NOT_IMPLEMENTED
    Code
      try(adbc_statement_set_sql_query(stmt, ""))
    Output
      LogStatementSetSqlQuery()
      Error in adbc_statement_set_sql_query(stmt, "") : NOT_IMPLEMENTED
    Code
      try(adbc_statement_set_substrait_plan(stmt, raw()))
    Output
      LogStatementSetSubstraitPlan()
      Error in adbc_statement_set_substrait_plan(stmt, raw()) : NOT_IMPLEMENTED
    Code
      try(adbc_statement_bind(stmt, data.frame()))
    Output
      LogStatementBind()
      Error in adbc_statement_bind(stmt, data.frame()) : NOT_IMPLEMENTED
    Code
      try(adbc_statement_bind_stream(stmt, data.frame()))
    Output
      LogStatementBindStream()
      Error in adbc_statement_bind_stream(stmt, data.frame()) : NOT_IMPLEMENTED
    Code
      try(adbc_statement_cancel(stmt))
    Output
      LogStatementCancel()
      Error in adbc_statement_cancel(stmt) : NOT_IMPLEMENTED
    Code
      adbc_statement_release(stmt)
    Output
      LogStatementRelease()
    Code
      adbc_connection_release(con)
    Output
      LogConnectionRelease()
    Code
      adbc_database_release(db)
    Output
      LogDatabaseRelease()

# default read/write/execute_adbc() performs the correct calls

    Code
      db <- adbc_database_init(adbc_driver_log())
    Output
      LogDatabaseNew()
      LogDatabaseInit()
    Code
      try(read_adbc(db, "some sql"))
    Output
      LogConnectionNew()
      LogConnectionInit()
      LogStatementNew()
      LogStatementSetSqlQuery()
      LogStatementRelease()
      LogConnectionRelease()
      Error in adbc_statement_set_sql_query(stmt, query) : NOT_IMPLEMENTED
    Code
      try(execute_adbc(db, "some sql"))
    Output
      LogConnectionNew()
      LogConnectionInit()
      LogStatementNew()
      LogStatementSetSqlQuery()
      LogStatementRelease()
      LogConnectionRelease()
      Error in adbc_statement_set_sql_query(stmt, query) : NOT_IMPLEMENTED
    Code
      try(write_adbc(mtcars, db, "some_table"))
    Output
      LogConnectionNew()
      LogConnectionInit()
      LogStatementNew()
      LogStatementSetOption()
      LogStatementBindStream()
      LogStatementRelease()
      LogConnectionRelease()
      Error in adbc_statement_bind_stream(stmt, tbl) : NOT_IMPLEMENTED
    Code
      adbc_database_release(db)
    Output
      LogDatabaseRelease()

---

    Code
      db <- adbc_database_init(adbc_driver_log())
    Output
      LogDatabaseNew()
      LogDatabaseInit()
    Code
      con <- adbc_connection_init(db)
    Output
      LogConnectionNew()
      LogConnectionInit()
    Code
      try(read_adbc(con, "some sql"))
    Output
      LogStatementNew()
      LogStatementSetSqlQuery()
      LogStatementRelease()
      Error in adbc_statement_set_sql_query(stmt, query) : NOT_IMPLEMENTED
    Code
      try(execute_adbc(con, "some sql"))
    Output
      LogStatementNew()
      LogStatementSetSqlQuery()
      LogStatementRelease()
      Error in adbc_statement_set_sql_query(stmt, query) : NOT_IMPLEMENTED
    Code
      try(write_adbc(mtcars, con, "some_table"))
    Output
      LogStatementNew()
      LogStatementSetOption()
      LogStatementBindStream()
      LogStatementRelease()
      Error in adbc_statement_bind_stream(stmt, tbl) : NOT_IMPLEMENTED
    Code
      adbc_connection_release(con)
    Output
      LogConnectionRelease()
    Code
      adbc_database_release(db)
    Output
      LogDatabaseRelease()

# joiners work for databases, connections, and statements

    Code
      stmt <- local({
        db <- local_adbc(adbc_database_init(adbc_driver_log()))
        con <- local_adbc(adbc_connection_init(db))
        adbc_connection_join(con, db)
        expect_false(adbc_xptr_is_valid(db))
        stmt <- local_adbc(adbc_statement_init(con))
        adbc_statement_join(stmt, con)
        expect_false(adbc_xptr_is_valid(con))
        adbc_xptr_move(stmt)
      })
    Output
      LogDatabaseNew()
      LogDatabaseInit()
      LogConnectionNew()
      LogConnectionInit()
      LogStatementNew()
    Code
      adbc_statement_release(stmt)
    Output
      LogStatementRelease()
      LogConnectionRelease()
      LogDatabaseRelease()

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/helpers.R
\name{adbc_connection_join}
\alias{adbc_connection_join}
\alias{adbc_statement_join}
\title{Join the lifecycle of a unique parent to its child}
\usage{
adbc_connection_join(connection, database)

adbc_statement_join(statement, connection)
}
\arguments{
\item{connection}{A connection created with \code{\link[=adbc_connection_init]{adbc_connection_init()}}}

\item{database}{A database created with \code{\link[=adbc_database_init]{adbc_database_init()}}}

\item{statement}{A statement created with \code{\link[=adbc_statement_init]{adbc_statement_init()}}}
}
\value{
The input, invisibly.
}
\description{
It is occasionally useful to return a connection, statement, or stream
from a function that was created from a unique parent. These helpers
tie the lifecycle of a unique parent object to its child such that the
parent object is released predictably and immediately after the child.
These functions will invalidate all references to the previous R object.
}
\examples{
# Use local_adbc to ensure prompt cleanup on error;
# use join functions to return a single object that manages
# the lifecycle of all three.
stmt <- local({
  db <- local_adbc(adbc_database_init(adbc_driver_log()))

  con <- local_adbc(adbc_connection_init(db))
  adbc_connection_join(con, db)

  stmt <- local_adbc(adbc_statement_init(con))
  adbc_statement_join(stmt, con)

  adbc_xptr_move(stmt)
})

# Everything is released immediately when the last object is released
adbc_statement_release(stmt)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/adbcdrivermanager-package.R
\docType{package}
\name{adbcdrivermanager-package}
\alias{adbcdrivermanager}
\alias{adbcdrivermanager-package}
\title{adbcdrivermanager: 'Arrow' Database Connectivity ('ADBC') Driver Manager}
\description{
Provides a developer-facing interface to 'Arrow' Database Connectivity ('ADBC') for the purposes of driver development, driver testing, and building high-level database interfaces for users. 'ADBC' \url{https://arrow.apache.org/adbc/} is an API standard for database access libraries that uses 'Arrow' for result sets and query parameters.
}
\seealso{
Useful links:
\itemize{
  \item \url{https://arrow.apache.org/adbc/current/r/adbcdrivermanager/}
  \item \url{https://github.com/apache/arrow-adbc}
  \item Report bugs at \url{https://github.com/apache/arrow-adbc/issues}
}

}
\author{
\strong{Maintainer}: Dewey Dunnington \email{dewey@dunnington.ca} (\href{https://orcid.org/0000-0002-9415-4582}{ORCID})

Authors:
\itemize{
  \item Apache Arrow \email{dev@arrow.apache.org} [copyright holder]
}

Other contributors:
\itemize{
  \item Apache Software Foundation \email{dev@arrow.apache.org} [copyright holder]
}

}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/helpers.R
\name{with_adbc}
\alias{with_adbc}
\alias{local_adbc}
\title{Cleanup helpers}
\usage{
with_adbc(x, code)

local_adbc(x, .local_envir = parent.frame())
}
\arguments{
\item{x}{An ADBC database, ADBC connection, ADBC statement, or
nanoarrow_array_stream returned from calls to an ADBC function.}

\item{code}{Code to execute before cleaning up the input.}

\item{.local_envir}{The execution environment whose scope should be tied
to the input.}
}
\value{
\itemize{
\item \code{with_adbc()} returns the result of \code{code}
\item \code{local_adbc()} returns the input, invisibly.
}
}
\description{
Managing the lifecycle of databases, connections, and statements can
be complex and error-prone. The R objects that wrap the underlying ADBC
pointers will perform cleanup in the correct order if you rely on garbage
collection (i.e., do nothing and let the objects go out of scope); however
it is good practice to explicitly clean up these objects. These helpers
are designed to make explicit and predictable cleanup easy to accomplish.
}
\details{
Note that you can use \code{\link[=adbc_connection_join]{adbc_connection_join()}} and \code{\link[=adbc_statement_join]{adbc_statement_join()}}
to tie the lifecycle of the parent object to that of the child object.
These functions mark any previous references to the parent object as
released so you can still use local and with helpers to manage the parent
object before it is joined. Use \code{stream_join_parent = TRUE} in
\code{\link[=adbc_statement_execute_query]{adbc_statement_execute_query()}} to tie the lifecycle of a statement to
the output stream.
}
\examples{
# Using with_adbc():
with_adbc(db <- adbc_database_init(adbc_driver_void()), {
  with_adbc(con <- adbc_connection_init(db), {
    with_adbc(stmt <- adbc_statement_init(con), {
      # adbc_statement_set_sql_query(stmt, "SELECT * FROM foofy")
      # adbc_statement_execute_query(stmt)
      "some result"
    })
  })
})

# Using local_adbc_*() (works best within a function, test, or local())
local({
  db <- local_adbc(adbc_database_init(adbc_driver_void()))
  con <- local_adbc(adbc_connection_init(db))
  stmt <- local_adbc(adbc_statement_init(con))
  # adbc_statement_set_sql_query(stmt, "SELECT * FROM foofy")
  # adbc_statement_execute_query(stmt)
  "some result"
})

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/adbc.R
\name{adbc_statement_set_sql_query}
\alias{adbc_statement_set_sql_query}
\alias{adbc_statement_set_substrait_plan}
\alias{adbc_statement_prepare}
\alias{adbc_statement_get_parameter_schema}
\alias{adbc_statement_bind}
\alias{adbc_statement_bind_stream}
\alias{adbc_statement_execute_query}
\alias{adbc_statement_execute_schema}
\alias{adbc_statement_cancel}
\title{Statement methods}
\usage{
adbc_statement_set_sql_query(statement, query)

adbc_statement_set_substrait_plan(statement, plan)

adbc_statement_prepare(statement)

adbc_statement_get_parameter_schema(statement)

adbc_statement_bind(statement, values, schema = NULL)

adbc_statement_bind_stream(statement, stream, schema = NULL)

adbc_statement_execute_query(
  statement,
  stream = NULL,
  stream_join_parent = FALSE
)

adbc_statement_execute_schema(statement)

adbc_statement_cancel(statement)
}
\arguments{
\item{statement}{An \link[=adbc_statement_init]{adbc_statement}}

\item{query}{An SQL query as a string}

\item{plan}{A raw vector representation of a serialized Substrait plan.}

\item{values}{A \link[nanoarrow:as_nanoarrow_array]{nanoarrow_array} or object
that can be coerced to one.}

\item{schema}{A \link[nanoarrow:as_nanoarrow_schema]{nanoarrow_schema} or object
that can be coerced to one.}

\item{stream}{A \link[nanoarrow:as_nanoarrow_array_stream]{nanoarrow_array_stream}
or object that can be coerced to one.}

\item{stream_join_parent}{Use \code{TRUE} to invalidate \code{statement} and tie its
lifecycle to \code{stream}.}
}
\value{
\itemize{
\item \code{adbc_statement_set_sql_query()}, \code{adbc_statement_set_substrait_plan()},
\code{adbc_statement_prepare()}, \code{adbc_statement_bind()},
\code{adbc_statement_bind_stream()}, and \code{adbc_statement_execute_query()}
return \code{statement}, invisibly.
\item \code{adbc_statement_get_parameter_schema()} returns a
\link[nanoarrow:as_nanoarrow_schema]{nanoarrow_schema}.
}
}
\description{
Statement methods
}
\examples{
db <- adbc_database_init(adbc_driver_void())
con <- adbc_connection_init(db)
stmt <- adbc_statement_init(con)
# (not implemented by the void driver)
try(adbc_statement_set_sql_query(stmt, "some query"))

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/driver_monkey.R
\name{adbc_driver_monkey}
\alias{adbc_driver_monkey}
\title{Monkey see, monkey do!}
\usage{
adbc_driver_monkey()
}
\value{
An object of class 'adbc_driver_monkey'
}
\description{
A driver whose query results are set in advance.
}
\examples{
db <- adbc_database_init(adbc_driver_monkey())
con <- adbc_connection_init(db)
stmt <- adbc_statement_init(con, mtcars)
stream <- nanoarrow::nanoarrow_allocate_array_stream()
adbc_statement_execute_query(stmt, stream)
as.data.frame(stream$get_next())

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/error.R
\name{adbc_error_from_array_stream}
\alias{adbc_error_from_array_stream}
\title{Get extended error information from an array stream}
\usage{
adbc_error_from_array_stream(stream)
}
\arguments{
\item{stream}{A \link[nanoarrow:as_nanoarrow_array_stream]{nanoarrow_array_stream}}
}
\value{
\code{NULL} if stream was not created by a driver that supports
extended error information or a list whose first element is the
status code and second element is the \code{adbc_error} object. The
\code{acbc_error} must not be accessed if \code{stream} is explicitly released.
}
\description{
Get extended error information from an array stream
}
\examples{
db <- adbc_database_init(adbc_driver_monkey())
con <- adbc_connection_init(db)
stmt <- adbc_statement_init(con, mtcars)
stream <- nanoarrow::nanoarrow_allocate_array_stream()
adbc_statement_execute_query(stmt, stream)
adbc_error_from_array_stream(stream)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/adbc.R, R/options.R
\name{adbc_database_init}
\alias{adbc_database_init}
\alias{adbc_database_init_default}
\alias{adbc_database_release}
\alias{adbc_database_set_options}
\alias{adbc_database_get_option}
\alias{adbc_database_get_option_bytes}
\alias{adbc_database_get_option_int}
\alias{adbc_database_get_option_double}
\title{Databases}
\usage{
adbc_database_init(driver, ...)

adbc_database_init_default(driver, options = NULL, subclass = character())

adbc_database_release(database)

adbc_database_set_options(database, options)

adbc_database_get_option(database, option)

adbc_database_get_option_bytes(database, option)

adbc_database_get_option_int(database, option)

adbc_database_get_option_double(database, option)
}
\arguments{
\item{driver}{An \code{\link[=adbc_driver]{adbc_driver()}}.}

\item{...}{Driver-specific options. For the default method, these are
named values that are converted to strings.}

\item{options}{A named \code{character()} or \code{list()} whose values are converted
to strings.}

\item{subclass}{An extended class for an object so that drivers can specify
finer-grained control over behaviour at the R level.}

\item{database}{An \link[=adbc_database_init]{adbc_database}.}

\item{option}{A specific option name}
}
\value{
An object of class adbc_database
}
\description{
Databases
}
\examples{
adbc_database_init(adbc_driver_void())

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/helpers.R
\name{read_adbc}
\alias{read_adbc}
\alias{execute_adbc}
\alias{write_adbc}
\title{Read, write, and execute on ADBC connections}
\usage{
read_adbc(db_or_con, query, ..., bind = NULL)

execute_adbc(db_or_con, query, ..., bind = NULL)

write_adbc(
  tbl,
  db_or_con,
  target_table,
  ...,
  mode = c("default", "create", "append", "replace", "create_append"),
  temporary = FALSE,
  catalog_name = NULL,
  db_schema_name = NULL
)
}
\arguments{
\item{db_or_con}{An adbc_database or adbc_connection. If a database, a
connection will be opened. For \code{read_adbc()}, this connection will
be closed when the resulting stream has been released.}

\item{query}{An SQL query}

\item{...}{Passed to S3 methods.}

\item{bind}{A data.frame, nanoarrow_array, or nanoarrow_array_stream of
bind parameters or NULL to skip the bind/prepare step.}

\item{tbl}{A data.frame, \link[nanoarrow:as_nanoarrow_array]{nanoarrow_array},
or  \link[nanoarrow:as_nanoarrow_array_stream]{nanoarrow_array_stream}.}

\item{target_table}{A target table name to which \code{tbl} should be written.}

\item{mode}{One of \code{"create"}, \code{"append"}, \code{"replace"}, \code{"create_append"} (error if the schema
is not compatible or append otherwise), or \code{"default"} (use the \code{adbc.ingest.mode}
argument of \code{\link[=adbc_statement_init]{adbc_statement_init()}}). The default is \code{"default"}.}

\item{temporary}{Use TRUE to create a table as a temporary table.}

\item{catalog_name}{If not \code{NULL}, the catalog to create/locate the table in.}

\item{db_schema_name}{If not \code{NULL}, the schema to create/locate the table in.}
}
\value{
\itemize{
\item \code{read_adbc()}: A \link[nanoarrow:as_nanoarrow_array_stream]{nanoarrow_array_stream}
\item \code{execute_adbc()}: \code{db_or_con}, invisibly.
\item \code{write_adbc()}: \code{tbl}, invisibly.
}
}
\description{
These are convenience methods useful for testing connections. Note that
S3 dispatch is always on \code{db_or_con} (i.e., drivers may provide their own
implementations).
}
\examples{
# On a database, connections are opened and closed
db <- adbc_database_init(adbc_driver_log())
try(read_adbc(db, "some sql"))
try(execute_adbc(db, "some sql"))
try(write_adbc(mtcars, db, "some_table"))

# Also works on a connection
con <- adbc_connection_init(db)
try(read_adbc(con, "some sql"))
try(execute_adbc(con, "some sql"))
try(write_adbc(mtcars, con, "some_table"))

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/adbc.R
\name{adbc_connection_get_info}
\alias{adbc_connection_get_info}
\alias{adbc_connection_get_objects}
\alias{adbc_connection_get_table_schema}
\alias{adbc_connection_get_table_types}
\alias{adbc_connection_read_partition}
\alias{adbc_connection_commit}
\alias{adbc_connection_rollback}
\alias{adbc_connection_cancel}
\alias{adbc_connection_get_statistic_names}
\alias{adbc_connection_get_statistics}
\alias{adbc_connection_quote_identifier}
\alias{adbc_connection_quote_string}
\title{Connection methods}
\usage{
adbc_connection_get_info(connection, info_codes = NULL)

adbc_connection_get_objects(
  connection,
  depth = 0L,
  catalog = NULL,
  db_schema = NULL,
  table_name = NULL,
  table_type = NULL,
  column_name = NULL
)

adbc_connection_get_table_schema(connection, catalog, db_schema, table_name)

adbc_connection_get_table_types(connection)

adbc_connection_read_partition(connection, serialized_partition)

adbc_connection_commit(connection)

adbc_connection_rollback(connection)

adbc_connection_cancel(connection)

adbc_connection_get_statistic_names(connection)

adbc_connection_get_statistics(
  connection,
  catalog,
  db_schema,
  table_name,
  approximate = FALSE
)

adbc_connection_quote_identifier(connection, value, ...)

adbc_connection_quote_string(connection, value, ...)
}
\arguments{
\item{connection}{An \link[=adbc_connection_init]{adbc_connection}}

\item{info_codes}{A list of metadata codes to fetch, or NULL to fetch all.
Valid values are documented in the adbc.h header.}

\item{depth}{The level of nesting to display. If 0, display all levels. If 1,
display only catalogs (i.e., catalog_schemas will be null). If 2, display
only catalogs and schemas (i.e., db_schema_tables will be null). If 3,
display only catalogs, schemas, and tables.}

\item{catalog}{Only show tables in the given catalog. If NULL, do not filter
by catalog. If an empty string, only show tables without a catalog. May be
a search pattern.}

\item{db_schema}{Only show tables in the given database schema. If NULL, do
not filter by database schema. If an empty string, only show tables without
a database schema. May be a search pattern.}

\item{table_name}{Constrain an object or statistics query for a specific table.
If NULL, do not filter by name. May be a search pattern.}

\item{table_type}{Only show tables matching one of the given table types. If
NULL, show tables of any type. Valid table types can be fetched from
GetTableTypes. Terminate the list with a NULL entry.}

\item{column_name}{Only show columns with the given name. If NULL, do not
filter by name. May be a search pattern.}

\item{serialized_partition}{The partition descriptor.}

\item{approximate}{If \code{FALSE}, request exact values of statistics,
else allow for best-effort, approximate, or cached values. The database
may return approximate values regardless, as indicated in the result.
Requesting exact values may be expensive or unsupported.}

\item{value}{A string or identifier.}

\item{...}{Driver-specific options. For the default method, these are
named values that are converted to strings.}
}
\value{
\itemize{
\item \code{adbc_connection_get_info()}, \code{adbc_connection_get_objects()},
\code{adbc_connection_get_table_types()}, and \code{adbc_connection_read_partition()}
return a \link[nanoarrow:as_nanoarrow_array_stream]{nanoarrow_array_stream}.
\item \code{adbc_connection_get_table_schema()} returns a
\link[nanoarrow:as_nanoarrow_schema]{nanoarrow_schena}
\item \code{adbc_connection_commit()} and \code{adbc_connection_rollback()} return
\code{connection}, invisibly.
}
}
\description{
Connection methods
}
\examples{
db <- adbc_database_init(adbc_driver_void())
con <- adbc_connection_init(db)
# (not implemented by the void driver)
try(adbc_connection_get_info(con, 0))

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/driver_void.R
\name{adbc_driver_void}
\alias{adbc_driver_void}
\alias{adbc_driver}
\title{Create ADBC drivers}
\usage{
adbc_driver_void()

adbc_driver(
  x,
  entrypoint = NULL,
  ...,
  load_flags = adbc_load_flags(),
  subclass = character()
)
}
\arguments{
\item{x, entrypoint}{An ADBC driver may be defined either as an init function
or as an identifier with an entrypoint name. A driver init func
must be an external pointer to a DL_FUNC with the type
\code{AdbcDriverInitFunc} specified in the adbc.h header.}

\item{...}{Further key/value parameters to store with the (R-level) driver
object.}

\item{load_flags}{Integer flags generated by \code{\link[=adbc_load_flags]{adbc_load_flags()}}}

\item{subclass}{An optional subclass for finer-grained control of
behaviour at the R level.}
}
\value{
An object of class 'adbc_driver'
}
\description{
Creates the R object representation of an ADBC driver, which consists of a
name and an initializer function with an optional subclass to control
finer-grained behaviour at the R level.
}
\examples{
adbc_driver_void()

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/driver_log.R
\name{adbc_driver_log}
\alias{adbc_driver_log}
\title{Log calls to another driver}
\usage{
adbc_driver_log()
}
\value{
An object of class 'adbc_driver_log'
}
\description{
Useful for debugging or ensuring that certain calls occur during
initialization and/or cleanup. The current logging output should not
be considered stable and may change in future releases.
}
\examples{
drv <- adbc_driver_log()
db <- adbc_database_init(drv, key = "value")
con <- adbc_connection_init(db, key = "value")
stmt <- adbc_statement_init(con, key = "value")
try(adbc_statement_execute_query(stmt))
adbc_statement_release(stmt)
adbc_connection_release(con)
adbc_database_release(db)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/adbc.R, R/options.R
\name{adbc_statement_init}
\alias{adbc_statement_init}
\alias{adbc_statement_init_default}
\alias{adbc_statement_release}
\alias{adbc_statement_set_options}
\alias{adbc_statement_get_option}
\alias{adbc_statement_get_option_bytes}
\alias{adbc_statement_get_option_int}
\alias{adbc_statement_get_option_double}
\title{Statements}
\usage{
adbc_statement_init(connection, ...)

adbc_statement_init_default(connection, options = NULL, subclass = character())

adbc_statement_release(statement)

adbc_statement_set_options(statement, options)

adbc_statement_get_option(statement, option)

adbc_statement_get_option_bytes(statement, option)

adbc_statement_get_option_int(statement, option)

adbc_statement_get_option_double(statement, option)
}
\arguments{
\item{connection}{An \link[=adbc_connection_init]{adbc_connection}}

\item{...}{Driver-specific options. For the default method, these are
named values that are converted to strings.}

\item{options}{A named \code{character()} or \code{list()} whose values are converted
to strings.}

\item{subclass}{An extended class for an object so that drivers can specify
finer-grained control over behaviour at the R level.}

\item{statement}{An \link[=adbc_statement_init]{adbc_statement}}

\item{option}{A specific option name}
}
\value{
An object of class 'adbc_statement'
}
\description{
Statements
}
\examples{
db <- adbc_database_init(adbc_driver_void())
con <- adbc_connection_init(db)
adbc_statement_init(con)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/adbc.R, R/options.R
\name{adbc_connection_init}
\alias{adbc_connection_init}
\alias{adbc_connection_init_default}
\alias{adbc_connection_release}
\alias{adbc_connection_set_options}
\alias{adbc_connection_get_option}
\alias{adbc_connection_get_option_bytes}
\alias{adbc_connection_get_option_int}
\alias{adbc_connection_get_option_double}
\title{Connections}
\usage{
adbc_connection_init(database, ...)

adbc_connection_init_default(database, options = NULL, subclass = character())

adbc_connection_release(connection)

adbc_connection_set_options(connection, options)

adbc_connection_get_option(connection, option)

adbc_connection_get_option_bytes(connection, option)

adbc_connection_get_option_int(connection, option)

adbc_connection_get_option_double(connection, option)
}
\arguments{
\item{database}{An \link[=adbc_database_init]{adbc_database}.}

\item{...}{Driver-specific options. For the default method, these are
named values that are converted to strings.}

\item{options}{A named \code{character()} or \code{list()} whose values are converted
to strings.}

\item{subclass}{An extended class for an object so that drivers can specify
finer-grained control over behaviour at the R level.}

\item{connection}{An \link[=adbc_connection_init]{adbc_connection}}

\item{option}{A specific option name}
}
\value{
An object of class 'adbc_connection'
}
\description{
Connections
}
\examples{
db <- adbc_database_init(adbc_driver_void())
adbc_connection_init(db)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/driver_void.R
\name{adbc_load_flags}
\alias{adbc_load_flags}
\title{Driver search/load options}
\usage{
adbc_load_flags(
  search_env = TRUE,
  search_user = TRUE,
  search_system = TRUE,
  allow_relative_paths = TRUE
)
}
\arguments{
\item{search_env}{Search for manifest files in the directories specified in
the ADBC_DRIVER_PATH environment variable and (when installed with conda)
in the conda environment.}

\item{search_user}{Search for manifest files in the designated directory
for user ADBC driver installs.}

\item{search_system}{Search for manifest files in the designtaed directory
for system ADBC driver installs.}

\item{allow_relative_paths}{Allow shared objects to be specified relative to
the current working directory.}
}
\value{
An integer flag value for use in \code{adbc_driver()}
}
\description{
Options that indicate where to look for driver manifests. Manifests
(.toml files) can be installed at the system level, the user level,
in location(s) specified by the ADBC_DRIVER_PATH environment variable,
and/or in a conda environment. See the ADBC documentation for details
regarding the locations of the user and system paths on various platforms.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/driver_void.R
\name{adbc_driver_load}
\alias{adbc_driver_load}
\title{Low-level driver loader}
\usage{
adbc_driver_load(
  x,
  entrypoint,
  version,
  driver,
  error,
  load_flags = adbc_load_flags(),
  additional_search_path_list = NULL
)
}
\arguments{
\item{x, entrypoint}{An ADBC driver may be defined either as an init function
or as an identifier with an entrypoint name. A driver init func
must be an external pointer to a DL_FUNC with the type
\code{AdbcDriverInitFunc} specified in the adbc.h header.}

\item{version}{The version number corresponding to the \code{driver} supplied}

\item{driver}{An external pointer to an \code{AdbcDriver}}

\item{error}{An external pointer to an \code{AdbcError} or NULL}

\item{load_flags}{Integer flags generated by \code{\link[=adbc_load_flags]{adbc_load_flags()}}}

\item{additional_search_path_list}{A path list of additional locations to search for driver manifests}
}
\value{
An integer ADBC status code
}
\description{
Most users should use \code{\link[=adbc_driver]{adbc_driver()}}; however, this function may be used
to allow other libraries (e.g., GDAL) to access the driver loader.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils.R
\name{adbc_xptr_move}
\alias{adbc_xptr_move}
\alias{adbc_xptr_is_valid}
\title{Low-level pointer details}
\usage{
adbc_xptr_move(x, check_child_count = TRUE)

adbc_xptr_is_valid(x)
}
\arguments{
\item{x}{An 'adbc_database', 'adbc_connection', 'adbc_statement', or
'nanoarrow_array_stream'}

\item{check_child_count}{Ensures that \code{x} has a zero child count before
performing the move. This should almost always be \code{TRUE}.}
}
\value{
\itemize{
\item \code{adbc_xptr_move()}: A freshly-allocated R object identical to \code{x}
\item \code{adbc_xptr_is_valid()}: Returns FALSE if the ADBC object pointed to by \code{x}
has been invalidated.
}
}
\description{
\itemize{
\item \code{adbc_xptr_move()} allocates a fresh R object and moves all values pointed
to by \code{x} into it. The original R object is invalidated by zeroing its
content. This is useful when returning from a function where
\link[=with_adbc]{lifecycle helpers} were used to manage the original
object.
\item \code{adbc_xptr_is_valid()} provides a means by which to test for an invalidated
pointer.
}
}
\examples{
db <- adbc_database_init(adbc_driver_void())
adbc_xptr_is_valid(db)
db_new <- adbc_xptr_move(db)
adbc_xptr_is_valid(db)
adbc_xptr_is_valid(db_new)

}
